from  __future__ import annotations

import inspect
import logging
from types import MethodType

try:
    from PyQt6.QtCore import Qt
    from PyQt6.QtGui import QPixmap, QPainter, QPalette
    from PyQt6.QtWidgets import (
        QWidget, QVBoxLayout, QLabel, QScrollArea,
        QSizePolicy, QToolButton, QHBoxLayout, QPushButton, QMainWindow, QGraphicsView, QGraphicsScene, QGraphicsPixmapItem
    )
    from qtconsole.inprocess import QtInProcessKernelManager
    from qtconsole.rich_jupyter_widget import RichJupyterWidget
except ImportError as e:
    logging.debug("RDRCaseViewer is not available. GUI features will not work. "
                  "Make sure you have PyQt6 installed if you want to use the GUI features.")
    raise ImportError("PyQt6 is required for the GUI features. Please install it using 'pip install PyQt6'") from e

from typing_extensions import Optional, Any, List, Dict, Callable

from ..datastructures.dataclasses import CaseQuery
from ..datastructures.enums import PromptFor
from .template_file_creator import TemplateFileCreator
from ..utils import is_iterable, contains_return_statement, encapsulate_code_lines_into_a_function
from .object_diagram import generate_object_graph


class ImageViewer(QGraphicsView):
    def __init__(self):
        super().__init__()
        self.setScene(QGraphicsScene(self))
        self.setDragMode(QGraphicsView.ScrollHandDrag)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QGraphicsView.AnchorUnderMouse)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        self.pixmap_item = None

        self._zoom = 0

    def update_image(self, image_path: str):
        if self.pixmap_item is None:
            self.pixmap_item = QGraphicsPixmapItem()
            self.scene().addItem(self.pixmap_item)
        pixmap = QPixmap(image_path)
        self.pixmap_item.setPixmap(pixmap)
        self.setSceneRect(self.pixmap_item.boundingRect())

    def wheelEvent(self, event):
        # Zoom in or out with Ctrl + mouse wheel
        if event.modifiers() == Qt.ControlModifier:
            angle = event.angleDelta().y()
            factor = 1.2 if angle > 0 else 0.8

            self._zoom += 1 if angle > 0 else -1
            if self._zoom > 10:  # max zoom in limit
                self._zoom = 10
                return
            if self._zoom < -10:  # max zoom out limit
                self._zoom = -10
                return

            self.scale(factor, factor)
        else:
            super().wheelEvent(event)


class BackgroundWidget(QWidget):
    def __init__(self, image_path, parent=None):
        super().__init__(parent)
        self.pixmap = QPixmap(image_path)

        # Layout for buttons
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(20, 20, 20, 20)
        self.layout.setSpacing(10)

        accept_btn = QPushButton("Accept")
        accept_btn.setStyleSheet("background-color: #4CAF50; color: white;")  # Green button
        edit_btn = QPushButton("Edit")
        edit_btn.setStyleSheet("background-color: #2196F3; color: white;")  # Blue button

        self.layout.addWidget(accept_btn)
        self.layout.addWidget(edit_btn)
        self.layout.addStretch()  # Push buttons to top

    def paintEvent(self, event):
        painter = QPainter(self)
        if not self.pixmap.isNull():
            # Calculate the vertical space used by buttons
            button_area_height = 0
            for i in range(self.layout.count()):
                item = self.layout.itemAt(i)
                if item.widget():
                    button_area_height += item.widget().height() + self.layout.spacing()

            remaining_height = self.height() - button_area_height
            if remaining_height <= 0:
                return  # No space to draw

            # Scale image to the remaining area (width, height)
            scaled = self.pixmap.scaled(
                self.width(),
                remaining_height,
                Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation
            )

            x = (self.width() - scaled.width()) // 2

            # Draw the image starting just below the buttons
            painter.drawPixmap(x, button_area_height + 20, scaled)

    def resizeEvent(self, event):
        self.update()  # Force repaint on resize
        super().resizeEvent(event)


class CollapsibleBox(QWidget):
    def __init__(self, title="", parent=None, viewer: Optional[RDRCaseViewer]=None, chain_name: Optional[str] = None,
                 main_obj: Optional[Dict[str, Any]] = None):
        super().__init__(parent)

        self.viewer = viewer
        self.chain_name = chain_name
        self.main_obj = main_obj
        self.toggle_button = QToolButton(checkable=True, checked=False)
        self.toggle_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self.toggle_button.setArrowType(Qt.ArrowType.RightArrow)
        self.toggle_button.clicked.connect(self.toggle)
        self.toggle_button.setStyleSheet("""
            QToolButton {
                border: none;
                font-weight: bold;
                color: #FFA07A; /* Light orange */
            }
        """)
        self.title_label = QLabel(title)
        self.title_label.setTextFormat(Qt.TextFormat.RichText)  # Enable HTML rendering
        self.title_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        self.title_label.setStyleSheet("QLabel { padding: 1px; color: #FFA07A; }")

        self.content_area = QWidget()
        self.content_area.setVisible(False)
        self.content_layout = QVBoxLayout(self.content_area)
        self.content_layout.setContentsMargins(15, 2, 0, 2)
        self.content_layout.setSpacing(2)

        layout = QVBoxLayout(self)
        header_layout = QHBoxLayout()
        header_layout.addWidget(self.toggle_button)
        header_layout.addWidget(self.title_label)
        layout.addLayout(header_layout)
        layout.addWidget(self.content_area)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

    def toggle(self):
        is_expanded = self.toggle_button.isChecked()
        self.update_object_diagram(is_expanded)
        self.toggle_button.setArrowType(
            Qt.ArrowType.DownArrow if is_expanded else Qt.ArrowType.RightArrow
        )
        self.content_area.setVisible(is_expanded)

        self.update_object_diagram(is_expanded)

        # toggle children
        if not is_expanded:
            for i in range(self.content_layout.count()):
                item = self.content_layout.itemAt(i)
                if isinstance(item.widget(), CollapsibleBox):
                    item.widget().toggle_button.setChecked(False)
                    item.widget().toggle()

    def update_object_diagram(self, is_expanded):
        if is_expanded and self.viewer is not None:
            self.viewer.included_attrs.append(self.chain_name)
            main_obj_name = self.chain_name.split('.')[0]
            main_obj = self.main_obj.get(main_obj_name)
            self.viewer.update_object_diagram(
                main_obj, main_obj_name
            )
        elif self.viewer is not None and self.chain_name in self.viewer.included_attrs:
            for name in self.viewer.included_attrs:
                if name.startswith(self.chain_name):
                    self.viewer.included_attrs.remove(name)
            main_obj_name = self.chain_name.split('.')[0]
            main_obj = self.main_obj.get(main_obj_name)
            self.viewer.update_object_diagram(
                main_obj, main_obj_name)

    def add_widget(self, widget):
        self.content_layout.addWidget(widget)

    def adjust_size_recursive(self):
        # Trigger resize
        self.adjustSize()

        # Traverse upwards to main window and call adjustSize on it too
        parent = self.parent()
        while parent:
            if isinstance(parent, QWidget):
                parent.layout().activate()  # Force layout refresh
                parent.adjustSize()
            elif isinstance(parent, QScrollArea):
                parent.widget().adjustSize()
                parent.viewport().update()
            if isinstance(parent, BackgroundWidget):
                parent.update()
                parent.updateGeometry()
                parent.repaint()
            if parent.parent() is None:
                top_window = parent.window()  # The main top-level window
                top_window.updateGeometry()
                top_window.repaint()
            parent = parent.parent()


def python_colored_repr(value):
    if isinstance(value, str):
        return f'<span style="color:#90EE90;">"{value}"</span>'
    elif isinstance(value, (int, float)):
        return f'<span style="color:#ADD8E6;">{value}</span>'
    elif isinstance(value, bool) or value is None:
        return f'<span style="color:darkorange;">{value}</span>'
    elif isinstance(value, type):
        return f'<span style="color:#C1BCBB;">{{{value.__name__}}}</span>'
    elif callable(value):
        return ''
    else:
        try:
            return f'<span style="color:white;">{repr(value)}</span>'
        except Exception as e:
            return f'<span style="color:red;">&lt;error: {e}&gt;</span>'


def style(text, color=None, font_size=None, font_weight=None):
    s = '<span style="'
    if color:
        s += f'color:{color_name_to_html(color)};'
    if font_size:
        s += f'font-size:{font_size}px;'
    if font_weight:
        s += f'font-weight:{font_weight};'
    s += '">'
    s += text
    s += '</span>'
    return s


def color_name_to_html(color_name):
    single_char_to_name = {
        'r': 'red',
        'g': 'green',
        'b': 'blue',
        'o': 'orange',
    }
    color_map = {
        'red': '#d6336c',
        'green': '#2eb872',
        'blue': '#007acc',
        'orange': '#FFA07A',
    }
    if len(color_name) == 1:
        color_name = single_char_to_name.get(color_name, color_name)
    return color_map.get(color_name.lower(), color_name)  # Default to the name itself if not found


class RDRCaseViewer(QMainWindow):
    case_query: Optional[CaseQuery] = None
    prompt_for: Optional[PromptFor] = None
    code_to_modify: Optional[str] = None
    template_file_creator: Optional[TemplateFileCreator] = None
    code_lines: Optional[List[str]] = None
    included_attrs: Optional[List[str]] = None
    main_obj: Optional[Dict[str, Any]] = None
    user_input: Optional[str] = None
    attributes_widget: Optional[QWidget] = None
    save_function: Optional[Callable[str], None] = None


    def __init__(self, parent=None, save_file: Optional[str] = None):
        super().__init__(parent)
        self.save_file = save_file

        self.setWindowTitle("RDR Case Viewer")

        self.setBaseSize(1600, 600)  # or your preferred initial size

        main_widget = QWidget()
        self.setCentralWidget(main_widget)
        self.setStyleSheet("background-color: #333333;")
        main_widget.setStyleSheet("background-color: #333333;")

        main_layout = QHBoxLayout(main_widget)  # Horizontal layout to split window

        # === Left: Attributes ===
        self._create_attribute_widget()

        # === Middle: Ipython & Action buttons ===
        middle_widget = QWidget()
        self.middle_widget_layout = QVBoxLayout(middle_widget)

        self.title_label = self.create_label_widget(style(style(f"Welcome to {style('RDRViewer', 'b')} "
                                                                f"{style('App', 'g')}")
                                                          , 'o', 28, 'bold'))

        self.buttons_widget = self.create_buttons_widget()

        self.ipython_console = IPythonConsole(parent=self)

        self.middle_widget_layout.addWidget(self.title_label)
        self.middle_widget_layout.addWidget(self.ipython_console)
        self.middle_widget_layout.addWidget(self.buttons_widget)

        # === Right: Object Diagram ===
        self.obj_diagram_viewer = ImageViewer()  # put your image path here

        # Add both to main layout
        main_layout.addWidget(self.attributes_widget, stretch=1)
        main_layout.addWidget(middle_widget, stretch=2)
        main_layout.addWidget(self.obj_diagram_viewer, stretch=2)

    def set_save_function(self, save_function: Callable[[str], None]) -> None:
        """
        Set the function to save the file.

        :param save_function: The function to save the file.
        """
        self.save_function = save_function
        self.save_btn.clicked.connect(lambda: self.save_function(self.save_file))

    def print(self, msg):
        """
        Print a message to the console.
        """
        self.ipython_console._append_plain_text(msg + '\n', True)

    def update_for_case_query(self, case_query: CaseQuery, title_txt: Optional[str] = None,
                              prompt_for: Optional[PromptFor] = None, code_to_modify: Optional[str] = None):
        self.case_query = case_query
        self.prompt_for = prompt_for
        self.code_to_modify = code_to_modify
        title_text = title_txt or ""
        case_attr_type = ', '.join([t.__name__ for t in case_query.core_attribute_type])
        case_attr_type = style(f"{case_attr_type}", 'g', 28, 'bold')
        case_name = style(f"{case_query.name}", 'b', 28, 'bold')
        title_text = style(f"{title_text} {case_name} of type {case_attr_type}", 'o', 28, 'bold')
        self.update_for_object(case_query.case, case_query.case_name, case_query.scope, title_text)

    def update_for_object(self, obj: Any, name: str, scope: Optional[dict] = None,
                          title_text: Optional[str] = None):
        self.update_main_obj(obj, name)
        title_text = title_text or style(f"{name}", 'o', 28, 'bold')
        scope = scope or {}
        scope.update({name: obj})
        self.update_object_diagram(obj, name)
        self.update_attribute_layout(obj, name)
        self.title_label.setText(title_text)
        self.ipython_console.update_namespace(scope)

    def update_main_obj(self, obj, name):
        self.main_obj = {name: obj}
        self.included_attrs = []
        self.user_input = None

    def update_object_diagram(self, obj: Any, name: str):
        self.included_attrs = self.included_attrs or []
        graph = generate_object_graph(obj, name, included_attrs=self.included_attrs)
        graph.render("object_diagram", view=False, format='svg')
        self.obj_diagram_viewer.update_image("object_diagram.svg")

    def _create_attribute_widget(self):
        main_widget = QWidget()
        main_layout = QVBoxLayout(main_widget)

        buttons_widget = QWidget()
        buttons_layout = QHBoxLayout(buttons_widget)

        expand_btn = QPushButton("Expand All")
        expand_btn.clicked.connect(self.expand_all)
        expand_btn.setStyleSheet(f"background-color: white; color: black;")  # Green button
        buttons_layout.addWidget(expand_btn)

        collapse_btn = QPushButton("Collapse All")
        collapse_btn.clicked.connect(self.collapse_all)
        collapse_btn.setStyleSheet(f"background-color: white; color: black;")  # Green button
        buttons_layout.addWidget(collapse_btn)

        main_layout.addWidget(buttons_widget)

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        attr_widget = QWidget()
        self.attr_widget_layout = QVBoxLayout(attr_widget)
        self.attr_widget_layout.setSpacing(2)
        self.attr_widget_layout.setContentsMargins(6, 6, 6, 6)
        attr_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        scroll.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        scroll.setWidget(attr_widget)
        main_layout.addWidget(scroll)
        self.attributes_widget = main_widget

    def expand_all(self):
        # Expand all collapsible boxes
        for i in range(self.attr_widget_layout.count()):
            item = self.attr_widget_layout.itemAt(i)
            if isinstance(item.widget(), CollapsibleBox):
                self.expand_collapse_all(item.widget(), expand=True)

    def collapse_all(self):
        # Collapse all collapsible boxes
        for i in range(self.attr_widget_layout.count()):
            item = self.attr_widget_layout.itemAt(i)
            if isinstance(item.widget(), CollapsibleBox):
                self.expand_collapse_all(item.widget(), expand=False)

    def expand_collapse_all(self, widget, expand=True, curr_depth=0, max_depth=2):
        if expand and curr_depth < max_depth:
            widget.toggle_button.setChecked(expand)
            widget.toggle()
            # do it for recursive children
            for i in range(widget.content_layout.count()):
                item = widget.content_layout.itemAt(i)
                if isinstance(item.widget(), CollapsibleBox):
                    self.expand_collapse_all(item.widget(), expand=True, curr_depth=curr_depth + 1, max_depth=max_depth)



    def create_buttons_widget(self):
        button_widget = QWidget()
        button_widget_layout = QVBoxLayout(button_widget)

        row_1_button_widget = QWidget()
        row_1_button_widget_layout = QHBoxLayout(row_1_button_widget)
        row_2_button_widget = QWidget()
        row_2_button_widget_layout = QHBoxLayout(row_2_button_widget)
        button_widget_layout.addWidget(row_1_button_widget)
        button_widget_layout.addWidget(row_2_button_widget)

        accept_btn = QPushButton("Accept")
        accept_btn.clicked.connect(self._accept)
        accept_btn.setStyleSheet(f"background-color: {color_name_to_html('g')}; color: white;")  # Green button

        edit_btn = QPushButton("Edit")
        edit_btn.clicked.connect(self._edit)
        edit_btn.setStyleSheet(f"background-color: {color_name_to_html('o')}; color: white;")  # Orange button

        load_btn = QPushButton("Load")
        load_btn.clicked.connect(self._load)
        load_btn.setStyleSheet(f"background-color: {color_name_to_html('b')}; color: white;")  # Blue button

        self.save_btn = QPushButton("Save")
        self.save_btn.setStyleSheet(f"background-color: {color_name_to_html('b')}; color: white;")  # Blue button

        row_1_button_widget_layout.addWidget(edit_btn)
        row_1_button_widget_layout.addWidget(load_btn)
        row_1_button_widget_layout.addWidget(accept_btn)
        row_2_button_widget_layout.addWidget(self.save_btn)
        return button_widget

    def _accept(self):
        # close the window
        self.close()

    def _edit(self):
        self.template_file_creator = TemplateFileCreator(self.case_query, self.prompt_for, self.code_to_modify,
                                                         self.print)
        self.template_file_creator.edit()

    def _load(self):
        if not self.template_file_creator:
            return
        self.code_lines, updates = self.template_file_creator.load(self.template_file_creator.temp_file_path,
                                                                    self.template_file_creator.func_name,
                                                                    self.template_file_creator.print_func)
        self.ipython_console.kernel.shell.user_ns.update(updates)
        if self.code_lines is not None:
            self.user_input = encapsulate_code_lines_into_a_function(
                self.code_lines, self.template_file_creator.func_name,
                self.template_file_creator.function_signature,
                self.template_file_creator.func_doc, self.case_query)
            self.case_query.scope.update(updates)
        self.template_file_creator = None

    def update_attribute_layout(self, obj, name: str):
        # Clear the existing layout
        clear_layout(self.attr_widget_layout)

        # Re-add the collapsible box with the new object
        self.add_collapsible(name, obj, self.attr_widget_layout, 0, 3, chain_name=name)
        self.attr_widget_layout.addStretch()

    def create_label_widget(self, text):
        # Create a QLabel with rich text
        label = QLabel()
        label.setText(text)
        label.setAlignment(Qt.AlignCenter)  # Optional: center the text
        label.setWordWrap(True)  # Optional: allow wrapping if needed
        return label

    def add_attributes(self, obj, name, layout, current_depth=0, max_depth=3, chain_name=None):
        if current_depth > max_depth:
            return
        if isinstance(obj, dict):
            items = obj.items()
        elif isinstance(obj, (list, tuple, set)):
            items = enumerate(obj)
        else:
            methods = []
            attributes = []
            iterables = []
            for attr in dir(obj):
                if attr.startswith("_") or attr == "scope":
                    continue
                try:
                    value = getattr(obj, attr)
                    if callable(value):
                        methods.append((attr, value))
                        continue
                    elif is_iterable(value):
                        iterables.append((attr, value))
                        continue
                except Exception as e:
                    value = f"<error: {e}>"
                attributes.append((attr, value))
            items = attributes + iterables + methods
        chain_name = chain_name if chain_name is not None else name
        for attr, value in items:
            attr = f"{attr}"
            try:
                if is_iterable(value) or hasattr(value, "__dict__") and not inspect.isfunction(value):
                    self.add_collapsible(attr, value, layout, current_depth + 1, max_depth, chain_name=f"{chain_name}.{attr}")
                else:
                    self.add_non_collapsible(attr, value, layout)
            except Exception as e:
                err = QLabel(f"<b>{attr}</b>: <span style='color:red;'>&lt;error: {e}&gt;</span>")
                err.setTextFormat(Qt.TextFormat.RichText)
                layout.addWidget(err)

    def add_collapsible(self, attr, value, layout, current_depth, max_depth, chain_name=None):
        type_name = type(value) if not isinstance(value, type) else value
        collapsible = CollapsibleBox(
            f'<b><span style="color:#FFA07A;">{attr}</span></b> {python_colored_repr(type_name)}', viewer=self,
        chain_name=chain_name, main_obj=self.main_obj)
        self.add_attributes(value, attr, collapsible.content_layout, current_depth, max_depth, chain_name=chain_name)
        layout.addWidget(collapsible)

    def add_non_collapsible(self, attr, value, layout):
        type_name = type(value) if not isinstance(value, type) else value
        text = f'<b><span style="color:#FFA07A;">{attr}</span></b> {python_colored_repr(type_name)}: {python_colored_repr(value)}'
        item_label = QLabel()
        item_label.setTextFormat(Qt.TextFormat.RichText)
        item_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        item_label.setStyleSheet("QLabel { padding: 1px; color: #FFA07A; }")
        item_label.setText(text)
        item_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        layout.addWidget(item_label)


class IPythonConsole(RichJupyterWidget):
    def __init__(self, namespace=None, parent=None):
        super(IPythonConsole, self).__init__(parent)

        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel()
        self.kernel = self.kernel_manager.kernel
        self.kernel.gui = 'qt'
        self.command_log = []

        # Monkey patch its run_cell method
        def custom_run_cell(this, raw_cell, **kwargs):
            print(raw_cell)
            if contains_return_statement(raw_cell) and 'def ' not in raw_cell:
                if self.parent.template_file_creator and self.parent.template_file_creator.func_name in raw_cell:
                    self.command_log = self.parent.code_lines
                self.command_log.append(raw_cell)
                this.history_manager.store_inputs(line_num=this.execution_count, source=raw_cell)
                return None
            result = original_run_cell(raw_cell, **kwargs)
            if result.error_in_exec is None and result.error_before_exec is None:
                self.command_log.append(raw_cell)
            return result

        original_run_cell = self.kernel.shell.run_cell
        self.kernel.shell.run_cell = MethodType(custom_run_cell, self.kernel.shell)

        self.kernel_client = self.kernel_manager.client()
        self.kernel_client.start_channels()

        # Update the user namespace with your custom variables
        if namespace:
            self.update_namespace(namespace)

        # Set the underlying QTextEdit's palette
        palette = QPalette()
        self._control.setPalette(palette)

        # Override the stylesheet to force background and text colors
        self.style_sheet = '''\
    QPlainTextEdit, QTextEdit {
        background-color: %(bgcolor)s;
        background-clip: padding;
        color: %(fgcolor)s;
        selection-background-color: %(select)s;
    }
    .inverted {
        background-color: %(fgcolor)s;
        color: %(bgcolor)s;
    }
    .error { color: red; }
    .in-prompt-number { font-weight: bold; color: %(in_prompt_number_color)s} }
    .in-prompt { font-weight: bold; color: %(in_prompt_color)s }
    .out-prompt-number { font-weight: bold; color: %(out_prompt_number_color)s }
    .out-prompt { font-weight: bold; color: %(out_prompt_color)s }
'''%dict(
            bgcolor='#0b0d0b', fgcolor='#47d9cc', select="#555",
            in_prompt_number_color='lime', in_prompt_color='lime',
            out_prompt_number_color='red', out_prompt_color='red'
        )

        # Use a dark syntax style like monokai
        self.syntax_style = 'monokai'
        # self.set_default_style(colors='lightbg')

        self.exit_requested.connect(self.stop)

    def update_namespace(self, namespace):
        """
        Update the user namespace with new variables.
        """
        self.kernel.shell.user_ns.update(namespace)

    def execute(self, source=None, hidden=False, interactive=False):
        # Log the command before execution
        source = source if source is not None else self.input_buffer
        # self.command_log.append(source)
        super().execute(source, hidden, interactive)

    def stop(self):
        self.kernel_client.stop_channels()
        self.kernel_manager.shutdown_kernel()


def clear_layout(layout):
    while layout.count():
        item = layout.takeAt(0)
        widget = item.widget()
        if widget is not None:
            widget.setParent(None)
        elif item.layout() is not None:
            clear_layout(item.layout())



from __main__ import Robot
from __main__ import PhysicalObject
from typing import List
from typing_extensions import Union


def conditions_292338421353147015053324365450552788888(case):
    def conditions_for_robot_contained_objects_of_type_physical_object(case: Robot) -> bool:
        """Get conditions on whether it's possible to conclude a value for Robot.contained_objects  of type PhysicalObject."""
        return len(case.parts) > 0
    return conditions_for_robot_contained_objects_of_type_physical_object(case)


def conclusion_292338421353147015053324365450552788888(case):
    def robot_contained_objects_of_type_physical_object(case: Robot) -> List[PhysicalObject]:
        """Get possible value(s) for Robot.contained_objects  of type PhysicalObject."""
        contained_objects = []
        for part in case.parts:
            contained_objects.extend(part.contained_objects)
        return contained_objects
    return robot_contained_objects_of_type_physical_object(case)





import ast
import os
import shutil
import socket
import subprocess
import tempfile
from functools import cached_property
from textwrap import indent, dedent

from colorama import Fore, Style
from typing_extensions import Optional, Type, List, Callable, Tuple, Dict

from ..datastructures.case import Case
from ..datastructures.dataclasses import CaseQuery
from ..datastructures.enums import Editor, PromptFor
from ..utils import str_to_snake_case, get_imports_from_scope, make_list, typing_hint_to_str, \
    get_imports_from_types, extract_function_source, extract_imports


def detect_available_editor() -> Optional[Editor]:
    """
    Detect the available editor on the system.

    :return: The first found editor that is available on the system.
    """
    editor_env = os.environ.get("RDR_EDITOR")
    if editor_env:
        return Editor.from_str(editor_env)
    for editor in [Editor.Pycharm, Editor.Code, Editor.CodeServer]:
        if shutil.which(editor.value):
            return editor
    return None


def is_port_in_use(port: int = 8080) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def start_code_server(workspace):
    """
    Start the code-server in the given workspace.
    """
    filename = os.path.join(os.path.dirname(__file__), "start-code-server.sh")
    os.system(f"chmod +x {filename}")
    print(f"Starting code-server at {filename}")
    return subprocess.Popen(["/bin/bash", filename, workspace], stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, text=True)


FunctionData = Tuple[Optional[List[str]], Optional[Dict[str, Callable]]]


class TemplateFileCreator:
    """
    A class to create a rule template file for a given case and prompt for the user to edit it.
    """
    temp_file_path: Optional[str] = None
    """
    The path to the temporary file that is created for the user to edit.
    """
    port: int = int(os.environ.get("RDR_EDITOR_PORT", 8080))
    """
    The port to use for the code-server.
    """
    process: Optional[subprocess.Popen] = None
    """
    The process of the code-server.
    """
    all_code_lines: Optional[List[str]] = None
    """
    The list of all code lines in the function in the temporary file.
    """

    def __init__(self, case_query: CaseQuery, prompt_for: PromptFor,
                 code_to_modify: Optional[str] = None, print_func: Callable[[str], None] = print):
        self.print_func = print_func
        self.code_to_modify = code_to_modify
        self.prompt_for = prompt_for
        self.case_query = case_query
        self.output_type = self.get_output_type()
        self.user_edit_line = 0
        self.func_name: str = self.get_func_name(self.prompt_for, self.case_query)
        self.func_doc: str = self.get_func_doc()
        self.function_signature: str = self.get_function_signature()
        self.editor: Optional[Editor] = detect_available_editor()
        self.workspace: str = os.environ.get("RDR_EDITOR_WORKSPACE", os.path.dirname(self.case_query.scope['__file__']))
        self.temp_file_path: str = os.path.join(self.workspace, "edit_code_here.py")

    def get_output_type(self) -> List[Type]:
        """
        :return: The output type of the function as a list of types.
        """
        if self.prompt_for == PromptFor.Conditions:
            output_type = bool
        else:
            output_type = self.case_query.attribute_type
        return make_list(output_type) if output_type is not None else None

    def edit(self):
        if self.editor is None:
            self.print_func(
                f"{Fore.RED}ERROR:: No editor found. Please install PyCharm, VSCode or code-server.{Style.RESET_ALL}")
            return

        boilerplate_code = self.build_boilerplate_code()
        self.write_to_file(boilerplate_code)

        self.open_file_in_editor()

    def open_file_in_editor(self):
        """
        Open the file in the available editor.
        """
        if self.editor == Editor.Pycharm:
            subprocess.Popen(["pycharm", "--line", str(self.user_edit_line), self.temp_file_path],
                             stdout=subprocess.DEVNULL,
                             stderr=subprocess.DEVNULL)
        elif self.editor == Editor.Code:
            subprocess.Popen(["code", self.temp_file_path])
        elif self.editor == Editor.CodeServer:
            try:
                subprocess.check_output(["pgrep", "-f", "code-server"])
                # check if same port is in use
                if is_port_in_use(self.port):
                    self.print_func(f"Code-server is already running on port {self.port}.")
                else:
                    raise ValueError("Port is not in use.")
            except (subprocess.CalledProcessError, ValueError) as e:
                self.process = start_code_server(self.workspace)
            self.print_func(f"Open code-server in your browser at http://localhost:{self.port}?folder={self.workspace}")
        self.print_func(f"Edit the file: {Fore.MAGENTA}{self.temp_file_path}")

    def build_boilerplate_code(self):
        imports = self.get_imports()
        if self.function_signature is None:
            self.function_signature = self.get_function_signature()
        if self.func_doc is None:
            self.func_doc = self.get_func_doc()
        if self.code_to_modify is not None:
            body = indent(dedent(self.code_to_modify), '    ')
        else:
            body = "    # Write your code here\n    pass"
        boilerplate = f"""{imports}\n\n{self.function_signature}\n    \"\"\"{self.func_doc}\"\"\"\n{body}"""
        self.user_edit_line = imports.count('\n') + 6
        return boilerplate

    def get_function_signature(self) -> str:
        if self.func_name is None:
            self.func_name = self.get_func_name(self.prompt_for, self.case_query)
        output_type_hint = self.get_output_type_hint()
        func_args = self.get_func_args()
        return f"def {self.func_name}({func_args}){output_type_hint}:"

    def get_output_type_hint(self) -> str:
        """
        :return: A string containing the output type hint for the function.
        """
        output_type_hint = ""
        if self.prompt_for == PromptFor.Conditions:
            output_type_hint = " -> bool"
        elif self.prompt_for == PromptFor.Conclusion:
            output_type_hint = f" -> {self.case_query.attribute_type_hint}"
        return output_type_hint

    def get_func_args(self) -> str:
        """
        :return: A string containing the function arguments.
        """
        if self.case_query.is_function:
            func_args = {}
            for k, v in self.case_query.case.items():
                if (self.case_query.function_args_type_hints is not None
                        and k in self.case_query.function_args_type_hints):
                    func_args[k] = typing_hint_to_str(self.case_query.function_args_type_hints[k])[0]
                else:
                    func_args[k] = type(v).__name__ if not isinstance(v, type) else f"Type[{v.__name__}]"
            func_args = ', '.join([f"{k}: {v}" if str(v) not in ["NoneType", "None"] else str(k)
                                   for k, v in func_args.items()])
        else:
            func_args = f"case: {self.case_type.__name__}"
        return func_args

    def write_to_file(self, code: str):
        if self.temp_file_path is None:
            tmp = tempfile.NamedTemporaryFile(mode='w+', delete=False, suffix=".py",
                                              dir=self.workspace)
            tmp.write(code)
            tmp.flush()
            self.temp_file_path = tmp.name
            tmp.close()
        else:
            with open(self.temp_file_path, 'w+') as f:
                f.write(code)

    def get_imports(self):
        """
        :return: A string containing the imports for the function.
        """
        case_type_imports = []
        if self.case_query.is_function:
            for k, v in self.case_query.case.items():
                if (self.case_query.function_args_type_hints is not None
                        and k in self.case_query.function_args_type_hints):
                    hint_list = typing_hint_to_str(self.case_query.function_args_type_hints[k])[1]
                    for hint in hint_list:
                        hint_split = hint.split('.')
                        if len(hint_split) > 1:
                            case_type_imports.append(f"from {'.'.join(hint_split[:-1])} import {hint_split[-1]}")
                else:
                    if isinstance(v, type):
                        case_type_imports.append(f"from {v.__module__} import {v.__name__}")
                    elif hasattr(v, "__module__") and not v.__module__.startswith("__"):
                        case_type_imports.append(f"\nfrom {type(v).__module__} import {type(v).__name__}")
        else:
            case_type_imports.append(f"from {self.case_type.__module__} import {self.case_type.__name__}")
        if self.output_type is None:
            output_type_imports = [f"from typing_extensions import Any"]
        else:
            output_type_imports = get_imports_from_types(self.output_type)
            if len(self.output_type) > 1:
                output_type_imports.append("from typing_extensions import Union")
            if list in self.output_type:
                output_type_imports.append("from typing_extensions import List")
        imports = get_imports_from_scope(self.case_query.scope)
        imports = [i for i in imports if ("get_ipython" not in i)]
        imports.extend(case_type_imports)
        imports.extend([oti for oti in output_type_imports if oti not in imports])
        imports = set(imports)
        return '\n'.join(imports)

    @staticmethod
    def get_core_attribute_types(case_query: CaseQuery) -> List[Type]:
        """
        Get the core attribute types of the case query.

        :return: A list of core attribute types.
        """
        attr_types = [t for t in case_query.core_attribute_type if t.__module__ != "builtins" and t is not None
                      and t is not type(None)]
        return attr_types

    def get_func_doc(self) -> Optional[str]:
        """
        :return: A string containing the function docstring.
        """
        type_data = f" of type {' or '.join(map(lambda c: c.__name__, self.get_core_attribute_types(self.case_query)))}"
        if self.prompt_for == PromptFor.Conditions:
            return (f"Get conditions on whether it's possible to conclude a value"
                    f" for {self.case_query.name} {type_data}.")
        else:
            return f"Get possible value(s) for {self.case_query.name} {type_data}."

    @staticmethod
    def get_func_name(prompt_for, case_query) -> Optional[str]:
        func_name = ""
        if prompt_for == PromptFor.Conditions:
            func_name = f"{prompt_for.value.lower()}_for_"
        case_name = case_query.name.replace(".", "_")
        if case_query.is_function:
            func_name += case_name.replace(f"_{case_query.attribute_name}", "")
        else:
            func_name += case_name
            attribute_types = TemplateFileCreator.get_core_attribute_types(case_query)
            attribute_type_names = [t.__name__ for t in attribute_types]
            func_name += f"_of_type_{'_or_'.join(attribute_type_names)}"
        return str_to_snake_case(func_name)

    @cached_property
    def case_type(self) -> Type:
        """
        Get the type of the case object in the current scope.

        :return: The type of the case object.
        """
        case = self.case_query.scope['case']
        return case._obj_type if isinstance(case, Case) else type(case)

    @staticmethod
    def load(file_path: str, func_name: str, print_func: Callable = print) -> FunctionData:
        """
        Load the function from the given file path.

        :param file_path: The path to the file to load.
        :param func_name: The name of the function to load.
        :param print_func: The function to use for printing messages.
        :return: A tuple containing the function source code and the function object as a dictionary
        with the function name as the key and the function object as the value.
        """
        if not file_path:
            print_func(f"{Fore.RED}ERROR:: No file to load. Run %edit first.{Style.RESET_ALL}")
            return None, None

        with open(file_path, 'r') as f:
            source = f.read()

        tree = ast.parse(source)
        updates = {}
        for node in tree.body:
            if isinstance(node, ast.FunctionDef) and node.name == func_name:
                exec_globals = {}
                scope = extract_imports(tree=tree)
                updates.update(scope)
                exec(source, scope, exec_globals)
                user_function = exec_globals[func_name]
                updates[func_name] = user_function
                print_func(f"{Fore.BLUE}Loaded `{func_name}` function into user namespace.{Style.RESET_ALL}")
                break
        if updates:
            all_code_lines = extract_function_source(file_path,
                                                          [func_name],
                                                          join_lines=False)[func_name]
            return all_code_lines, updates
        else:
            print_func(f"{Fore.RED}ERROR:: Function `{func_name}` not found.{Style.RESET_ALL}")
            return None, None

    def __del__(self):
        if hasattr(self, 'process') and self.process is not None and self.process.poll() is None:
            self.process.terminate()  # Graceful shutdown
            self.process.wait()  # Ensure cleanup



import os
from unittest import TestCase

from ripple_down_rules.utils import extract_imports, make_set


class UtilsTestCase(TestCase):

    def test_extract_imports_from_file(self):
        # Test with a file that has imports
        file_path = "test_file.py"
        with open(file_path, "w") as f:
            f.write("import os\n")
            f.write("from module import function\n")
            f.write("from ripple_down_rules.utils import make_set\n")
            f.write("print('Hello World')\n")

        expected_scope = {"os": os, "make_set": make_set}
        actual_imports = extract_imports(file_path)
        self.assertEqual(expected_scope, actual_imports)

        # Clean up
        os.remove(file_path)



from __future__ import annotations

import os.path
from dataclasses import dataclass, field

from typing_extensions import List, Optional

from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.rdr import GeneralRDR


@dataclass(unsafe_hash=True)
class PhysicalObject:
    """
    A physical object is an object that can be contained in a container.
    """
    name: str
    contained_objects: List[PhysicalObject] = field(default_factory=list, hash=False)

@dataclass(unsafe_hash=True)
class Part(PhysicalObject):
    ...

@dataclass(unsafe_hash=True)
class Robot(PhysicalObject):
    parts: List[Part] = field(default_factory=list, hash=False)


part_a = Part(name="A")
part_b = Part(name="B")
part_c = Part(name="C")
robot = Robot("pr2", parts=[part_a])
part_a.contained_objects = [part_b]
part_b.contained_objects = [part_c]

case_query = CaseQuery(robot, "contained_objects", (PhysicalObject,), False)

load = True  # Set to True if you want to load an existing model, False if you want to create a new one.
if load and os.path.exists('./part_containment_rdr'):
    grdr = GeneralRDR.load('./', model_name='part_containment_rdr')
    grdr.ask_always = False # Set to True if you want to always ask the expert for a target value.
else:
    grdr = GeneralRDR(save_dir='./', model_name='part_containment_rdr')

grdr.fit_case(case_query)

print(grdr.classify(robot)['contained_objects'])
assert grdr.classify(robot)['contained_objects'] == {part_b}


import os

from typing_extensions import List, Any, Tuple, Type

from datasets import Species, Habitat
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import MultiClassRDR, SingleClassRDR, GeneralRDR
from ripple_down_rules.utils import make_set


def get_fit_scrdr(cases: List[Any], targets: List[Any], attribute_name: str = "species",
                  attribute_type: Type = Species,
                  expert_answers_dir: str = "test_expert_answers",
                  expert_answers_file: str = "scrdr_expert_answers_fit",
                  draw_tree: bool = False,
                  load_answers: bool = True,
                  save_answers: bool = False) -> Tuple[SingleClassRDR, List[CaseQuery]]:
    filename = os.path.join(os.getcwd(), expert_answers_dir, expert_answers_file)
    expert = Human(use_loaded_answers=load_answers)
    if load_answers:
        expert.load_answers(filename)

    targets = [None for _ in cases] if targets is None or len(targets) == 0 else targets
    scrdr = SingleClassRDR()
    case_queries = [CaseQuery(case, attribute_name, (attribute_type,), True, _target=target)
                    for case, target in zip(cases, targets)]
    scrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
    if save_answers:
        expert.save_answers(filename)
    for case_query in case_queries:
        cat = scrdr.classify(case_query.case)
        assert cat == case_query.target_value
    return scrdr, case_queries


def get_fit_mcrdr(cases: List[Any], targets: List[Any], attribute_name: str = "species",
                  attribute_type: Type = Species,
                  mutually_exclusive: bool = True,
                  expert_answers_dir: str = "test_expert_answers",
                  expert_answers_file: str = "mcrdr_expert_answers_stop_only_fit",
                  draw_tree: bool = False,
                  load_answers: bool = True,
                  save_answers: bool = False) -> MultiClassRDR:
    filename = os.path.join(os.getcwd(), expert_answers_dir, expert_answers_file)
    expert = Human(use_loaded_answers=load_answers)
    if load_answers:
        expert.load_answers(filename)
    targets = [None for _ in cases] if targets is None or len(targets) == 0 else targets
    mcrdr = MultiClassRDR()
    case_queries = [CaseQuery(case, attribute_name, (attribute_type,), mutually_exclusive, _target=target)
                    for case, target in zip(cases, targets)]
    mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
    if save_answers:
        expert.save_answers(filename)
    for case_query in case_queries:
        cat = mcrdr.classify(case_query.case)
        assert make_set(cat) == make_set(case_query.target_value)
    return mcrdr


def get_fit_grdr(cases: List[Any], targets: List[Any], expert_answers_dir: str = "test_expert_answers",
                 expert_answers_file: str = "grdr_expert_answers_fit", draw_tree: bool = False,
                 load_answers: bool = True,
                 save_answers: bool = False,
                 append: bool = False,
                 no_targets: bool = False) -> Tuple[GeneralRDR, List[dict]]:
    filename = os.path.join(os.path.dirname(__file__), '..', expert_answers_dir, expert_answers_file)
    expert = Human(use_loaded_answers=load_answers, append=append)
    if load_answers:
        expert.load_answers(filename)

    fit_scrdr, _ = get_fit_scrdr(cases, targets, draw_tree=False)

    grdr = GeneralRDR()
    grdr.add_rdr(fit_scrdr)

    n = 20
    true_targets = [get_habitat(x, t) for x, t in zip(cases[:n], targets[:n])]
    if no_targets:
        all_targets = [{'habitats': None} for i in range(n)]
    else:
        all_targets = true_targets
    case_queries = [CaseQuery(case, name, (Species,) if name == "species" else (Habitat,),
                              True if name == "species" else False, _target=target)
                    for case, targets in zip(cases[:n], all_targets)
                    for name, target in targets.items()]
    grdr.fit(case_queries, expert=expert,
             animate_tree=draw_tree)
    if save_answers:
        expert.save_answers(filename)
    for case, case_targets in zip(cases[:n], true_targets):
        cat = grdr.classify(case)
        for cat_name, cat_val in cat.items():
            if cat_name in case_targets:
                assert make_set(cat_val) == make_set(case_targets[cat_name])
    return grdr, true_targets


def get_habitat(x: Case, t: Category):
    habitat = set()
    if t == Species.mammal and x["aquatic"] == 0:
        habitat = {Habitat.land}
    elif t == Species.bird:
        habitat = {Habitat.land}
        if x["airborne"] == 1:
            habitat.update({Habitat.air})
        if x["aquatic"] == 1:
            habitat.update({Habitat.water})
    elif t == Species.fish:
        habitat = {Habitat.water}
    elif t == Species.molusc:
        habitat = {Habitat.land}
        if x["aquatic"] == 1:
            habitat.update({Habitat.water})
    if len(habitat) == 0:
        return {t.__class__.__name__.lower(): t}
    else:
        return {"habitats": habitat, t.__class__.__name__.lower(): t}



"""
This file contains decorators for the RDR (Ripple Down Rules) framework. Where each type of RDR has a decorator
that can be used with any python function such that this function can benefit from the incremental knowledge acquisition
of the RDRs.
"""
import os.path
from functools import wraps

from pyparsing.tools.cvt_pyparsing_pep8_names import camel_to_snake
from typing_extensions import Callable, Optional, Type, Tuple, Dict, Any, Self, get_type_hints, List, Union

from ripple_down_rules.datastructures.case import create_case, Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import Category
from ripple_down_rules.experts import Expert, Human
from ripple_down_rules.rdr import GeneralRDR, RippleDownRules
from ripple_down_rules.utils import get_method_args_as_dict, get_func_rdr_model_name, make_set, \
    get_method_class_if_exists, get_method_name, str_to_snake_case


class RDRDecorator:
    rdr: GeneralRDR

    def __init__(self, models_dir: str,
                 output_type: Tuple[Type],
                 mutual_exclusive: bool,
                 python_dir: Optional[str] = None,
                 output_name: str = "output_",
                 fit: bool = True,
                 expert: Optional[Expert] = None):
        """
        :param models_dir: The directory to save/load the RDR models.
        :param output_type: The type of the output. This is used to create the RDR model.
        :param mutual_exclusive: If True, the output types are mutually exclusive.
        :param python_dir: The directory to save the RDR model as a python file.
         If None, the RDR model will not be saved as a python file.
        :param output_name: The name of the output. This is used to create the RDR model.
        :param fit: If True, the function will be in fit mode. This means that the RDR will prompt the user for the
            correct output if the function's output is not in the RDR model. If False, the function will be in
            classification mode. This means that the RDR will classify the function's output based on the RDR model.
        :param expert: The expert that will be used to prompt the user for the correct output. If None, a Human
            expert will be used.
        :return: A decorator to use a GeneralRDR as a classifier that monitors and modifies the function's output.
        """
        self.rdr_models_dir = models_dir
        self.model_name: Optional[str] = None
        self.output_type = output_type
        self.parsed_output_type: List[Type] = []
        self.mutual_exclusive = mutual_exclusive
        self.rdr_python_path: Optional[str] = python_dir
        self.output_name = output_name
        self.fit: bool = fit
        self.expert: Optional[Expert] = expert
        self.load()

    def decorator(self, func: Callable) -> Callable:

        @wraps(func)
        def wrapper(*args, **kwargs) -> Optional[Any]:

            if len(self.parsed_output_type) == 0:
                self.parsed_output_type = self.parse_output_type(func, self.output_type, *args)
            if self.model_name is None:
                self.initialize_rdr_model_name_and_load(func)

            if self.fit:
                expert_answers_path = os.path.join(self.rdr_models_dir, self.model_name, "expert_answers")
                self.expert = self.expert or Human(answers_save_path=expert_answers_path)
                case_query = self.create_case_query_from_method(func, self.parsed_output_type,
                                                                self.mutual_exclusive, self.output_name,
                                                                *args, **kwargs)
                output = self.rdr.fit_case(case_query, expert=self.expert)
                return output[self.output_name]
            else:
                case, case_dict = self.create_case_from_method(func, self.output_name, *args, **kwargs)
                return self.rdr.classify(case)[self.output_name]

        return wrapper

    @staticmethod
    def create_case_query_from_method(func: Callable, output_type, mutual_exclusive: bool,
                                output_name: str = 'output_', *args, **kwargs) -> CaseQuery:
        """
        Create a CaseQuery from the function and its arguments.

        :param func: The function to create a case from.
        :param output_type: The type of the output.
        :param mutual_exclusive: If True, the output types are mutually exclusive.
        :param output_name: The name of the output in the case. Defaults to 'output_'.
        :param args: The positional arguments of the function.
        :param kwargs: The keyword arguments of the function.
        :return: A CaseQuery object representing the case.
        """
        output_type = make_set(output_type)
        case, case_dict = RDRDecorator.create_case_from_method(func, output_name, *args, **kwargs)
        scope = func.__globals__
        scope.update(case_dict)
        func_args_type_hints = get_type_hints(func)
        func_args_type_hints.update({output_name: Union[tuple(output_type)]})
        return CaseQuery(case, output_name, Union[tuple(output_type)],
                         mutual_exclusive, scope=scope,
                         is_function=True, function_args_type_hints=func_args_type_hints)

    @staticmethod
    def create_case_from_method(func: Callable, output_name: str = "output_", *args, **kwargs) -> Tuple[Case, Dict[str, Any]]:
        """
        Create a Case from the function and its arguments.

        :param func: The function to create a case from.
        :param output_name: The name of the output in the case. Defaults to 'output_'.
        :param args: The positional arguments of the function.
        :param kwargs: The keyword arguments of the function.
        :return: A Case object representing the case.
        """
        case_dict = get_method_args_as_dict(func, *args, **kwargs)
        func_output = func(*args, **kwargs)
        case_dict.update({output_name: func_output})
        case_name = get_func_rdr_model_name(func)
        return create_case(case_dict, obj_name=case_name, max_recursion_idx=3), case_dict

    def initialize_rdr_model_name_and_load(self, func: Callable) -> None:
        model_file_name = get_func_rdr_model_name(func, include_file_name=True)
        self.model_name = str_to_snake_case(model_file_name)
        self.load()

    @staticmethod
    def parse_output_type(func: Callable, output_type: Any, *args) -> List[Type]:
        parsed_output_type = []
        for ot in make_set(output_type):
            if ot is Self:
                func_class = get_method_class_if_exists(func, *args)
                if func_class is not None:
                    parsed_output_type.append(func_class)
                else:
                    raise ValueError(f"The function {func} is not a method of a class,"
                                     f" and the output type is {Self}.")
            else:
                parsed_output_type.append(ot)
        return parsed_output_type

    def save(self):
        """
        Save the RDR model to the specified directory.
        """
        self.rdr.save(self.rdr_models_dir)

    def load(self):
        """
        Load the RDR model from the specified directory.
        """
        if self.model_name is not None and os.path.exists(os.path.join(self.rdr_models_dir, self.model_name)):
            self.rdr = GeneralRDR.load(self.rdr_models_dir, self.model_name)
        else:
            self.rdr = GeneralRDR(save_dir=self.rdr_models_dir, model_name=self.model_name)

    def update_from_python(self):
        """
        Update the RDR model from a python file.
        """
        self.rdr.update_from_python(self.rdr_models_dir, self.model_name)



from __future__ import annotations

import os

from .datastructures.dataclasses import CaseQuery
from sqlalchemy.orm import Session
from typing_extensions import Type, Optional, Callable, Any, Dict, TYPE_CHECKING

from .utils import get_func_rdr_model_path
from .utils import calculate_precision_and_recall

if TYPE_CHECKING:
    from .rdr import RippleDownRules


def is_matching(classifier: Callable[[Any], Any], case_query: CaseQuery, pred_cat: Optional[Dict[str, Any]] = None) -> bool:
    """
    :param classifier: The RDR classifier to check the prediction of.
    :param case_query: The case query to check.
    :param pred_cat: The predicted category.
    :return: Whether the classifier prediction is matching case_query target or not.
    """
    if case_query.target is None:
        return False
    if pred_cat is None:
        pred_cat = classifier(case_query.case)
    if not isinstance(pred_cat, dict):
        pred_cat = {case_query.attribute_name: pred_cat}
    target = {case_query.attribute_name: case_query.target_value}
    precision, recall = calculate_precision_and_recall(pred_cat, target)
    return all(recall) and all(precision)


def load_or_create_func_rdr_model(func, model_dir: str, rdr_type: Type[RippleDownRules],
                                  session: Optional[Session] = None, **rdr_kwargs) -> RippleDownRules:
    """
    Load the RDR model of the function if it exists, otherwise create a new one.

    :param func: The function to load the model for.
    :param model_dir: The directory where the model is stored.
    :param rdr_type: The type of the RDR model to load.
    :param session: The SQLAlchemy session to use.
    :param rdr_kwargs: Additional arguments to pass to the RDR constructor in the case of a new model.
    """
    model_path = get_func_rdr_model_path(func, model_dir)
    if os.path.exists(model_path):
        rdr = rdr_type.load(model_path)
        rdr.session = session
    else:
        rdr = rdr_type(session=session, **rdr_kwargs)
    return rdr



from unittest import TestCase, skip

from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import PromptFor
from ripple_down_rules.user_interface.prompt import UserPrompt
from test_rdr_world import World, Handle, Container, FixedConnection, PrismaticConnection


class TestPrompt(TestCase):

    @skip
    def test_prompt(self):
        user_prompt = UserPrompt()
        world = World()

        handle = Handle('h1', world=world)
        handle_2 = Handle('h2', world=world)
        container_1 = Container('c1', world=world)
        container_2 = Container('c2', world=world)
        connection_1 = FixedConnection(container_1, handle, world=world)
        connection_2 = PrismaticConnection(container_2, container_1, world=world)

        world.bodies = [handle, container_1, container_2, handle_2]
        world.connections = [connection_1, connection_2]
        case_query = CaseQuery(world, "views", (Handle,), False)
        user_prompt.prompt_user_about_case(case_query, prompt_for=PromptFor.Conclusion)


import os
import sys
from unittest import TestCase

from typing_extensions import List, Optional

from datasets import Habitat, Species
from datasets import load_zoo_dataset
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.datastructures.enums import MCRDRMode
from ripple_down_rules.experts import Human
from ripple_down_rules.rdr import SingleClassRDR, MultiClassRDR, GeneralRDR, RDRWithCodeWriter
from ripple_down_rules.utils import render_tree, make_set, extract_function_source
from test_helpers.helpers import get_fit_scrdr, get_fit_mcrdr, get_fit_grdr

try:
    from PyQt6.QtWidgets import QApplication
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
except ImportError as e:
    RDRCaseViewer = None
    QApplication = None


class TestRDR(TestCase):
    all_cases: List[Case]
    targets: List[str]
    case_queries: List[CaseQuery]
    test_results_dir: str = "./test_results"
    expert_answers_dir: str = "./test_expert_answers"
    generated_rdrs_dir: str = "./test_generated_rdrs"
    cache_file: str = f"{test_results_dir}/zoo_dataset.pkl"
    app: Optional[QApplication] = None
    viewer: Optional[RDRCaseViewer] = None
    use_gui: bool = False

    @classmethod
    def setUpClass(cls):
        # fetch dataset
        cls.all_cases, cls.targets = load_zoo_dataset(cache_file=cls.cache_file)
        cls.case_queries = [CaseQuery(case, "species", Species, True, _target=target)
                            for case, target in zip(cls.all_cases, cls.targets)]
        for test_dir in [cls.test_results_dir, cls.expert_answers_dir, cls.generated_rdrs_dir]:
            if not os.path.exists(test_dir):
                os.makedirs(test_dir)
        if RDRCaseViewer is not None and QApplication is not None and cls.use_gui:
            cls.app = QApplication(sys.argv)
            cls.viewer = RDRCaseViewer()

    def test_classify_scrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/scrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers, viewer=self.viewer)
        if use_loaded_answers:
            expert.load_answers(filename)

        scrdr = SingleClassRDR()
        cat = scrdr.fit_case(self.case_queries[0], expert=expert)
        self.assertEqual(cat, self.targets[0])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_scrdr(self):
        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets, draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_expert_answers_fit",
                                 load_answers=True)
        # render_tree(scrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + f"/scrdr")

    def test_save_load_scrdr(self):

        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets, draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_expert_answers_fit",
                                 load_answers=True)
        save_dir = self.generated_rdrs_dir
        model_name = scrdr.save(save_dir)
        assert os.path.exists(save_dir)
        assert os.path.exists(save_dir + f"/{model_name}/{scrdr.metadata_folder}")
        assert os.path.exists(save_dir
                              + f"/{model_name}/{scrdr.metadata_folder}/{model_name}.json")
        assert os.path.exists(save_dir + f"/__init__.py")
        assert os.path.exists(save_dir + f"/{model_name}/__init__.py")
        assert os.path.exists(save_dir + f"/{model_name}/{scrdr.generated_python_file_name}.py")
        assert os.path.exists(save_dir + f"/{model_name}/{scrdr.generated_python_defs_file_name}.py")

        scrdr_loaded = SingleClassRDR.load(save_dir, scrdr.generated_python_file_name)
        self.assertEqual(scrdr.start_rule.uid, scrdr_loaded.start_rule.uid)
        for case, target in zip(self.all_cases, self.targets):
            cat = scrdr_loaded.classify(case)
            self.assertEqual(cat, target)

    def test_expert_incremental_save(self):
        if not os.path.exists(os.path.join(self.test_results_dir, "expert_incremental_save/expert_answers.py")):
            return
            # os.remove(os.path.join(self.test_results_dir, "expert_incremental_save/expert_answers.py"))
        expert = Human(answers_save_path=f"{self.test_results_dir}/expert_incremental_save/expert_answers")
        cq = CaseQuery(self.all_cases[0], "species", Species, True)
        conclusion = expert.ask_for_conclusion(cq)
        assert os.path.exists(self.test_results_dir + "/expert_incremental_save/expert_answers.py")

    def test_scrdr_incremental_save_and_load(self):
        if os.path.exists(self.test_results_dir + "/scrdr_incremental_save/animal_species_scrdr"):
            scrdr = SingleClassRDR.load(f"{self.test_results_dir}/scrdr_incremental_save",
                                        "animal_species_scrdr")
            scrdr.ask_always = False
        else:
            scrdr = SingleClassRDR(save_dir=self.test_results_dir + "/scrdr_incremental_save")
            return
        expert = Human(answers_save_path=f"{self.test_results_dir}/scrdr_incremental_save/expert_answers")
        cq = CaseQuery(self.all_cases[0], "species", Species, True)
        scrdr.fit_case(cq, expert=expert)
        assert not os.path.exists(self.test_results_dir + "/scrdr_incremental_save/expert_answers.py")
        assert scrdr.classify(self.all_cases[0]) == self.targets[0]

    def test_fit_scrdr_with_no_targets(self):
        # Test with no targets
        draw = False
        scrdr, case_queries = get_fit_scrdr(self.all_cases[:20], [], draw_tree=draw,
                                            expert_answers_dir=self.expert_answers_dir,
                                            expert_answers_file="scrdr_expert_answers_fit_no_targets",
                                            load_answers=True,
                                            save_answers=False)
        if draw:
            render_tree(scrdr.start_rule, use_dot_exporter=True,
                        filename=self.test_results_dir + f"/scrdr_no_targets")

    def test_write_scrdr_no_targets_to_python_file(self):
        # Test with no targets
        scrdr, case_queries = get_fit_scrdr(self.all_cases[:20], [], draw_tree=False,
                                            expert_answers_dir=self.expert_answers_dir,
                                            expert_answers_file="scrdr_expert_answers_fit_no_targets",
                                            load_answers=True, save_answers=False)
        model_dir = self.generated_rdrs_dir + '/scrdr_no_targets'
        os.makedirs(model_dir, exist_ok=True)
        scrdr._write_to_python(model_dir)
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(model_dir)
        for case_query, target in zip(case_queries, self.targets):
            cat = classify_species_scrdr(case_query.case)
            self.assertEqual(cat, target)

    def test_fit_mcrdr_with_no_targets(self):
        # Test with no targets
        mcrdr = get_fit_mcrdr(self.all_cases[:20], [], draw_tree=False,
                              expert_answers_dir=self.expert_answers_dir,
                              expert_answers_file="mcrdr_expert_answers_fit_no_targets",
                              load_answers=True,
                              save_answers=False)
        # render_tree(mcrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + f"/mcrdr_no_targets")
        for case, target in zip(self.all_cases[:20], self.targets[:20]):
            cat = mcrdr.classify(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_write_mcrdr_no_targets_to_python_file(self):
        # Test with no targets
        mcrdr = get_fit_mcrdr(self.all_cases[:20], [], draw_tree=False,
                              expert_answers_dir=self.expert_answers_dir,
                              expert_answers_file="mcrdr_expert_answers_fit_no_targets",
                              load_answers=True, save_answers=False)
        model_dir = self.generated_rdrs_dir + '/mcrdr_no_targets'
        os.makedirs(model_dir, exist_ok=True)
        mcrdr._write_to_python(model_dir)
        classify_species_mcrdr = mcrdr.get_rdr_classifier_from_python_file(model_dir)
        for case, target in zip(self.all_cases[:20], self.targets[:20]):
            cat = classify_species_mcrdr(case)
            self.assertEqual(make_set(cat), make_set(target))

    # @skip("Test is not implemented yet")
    def test_fit_grdr_with_no_targets(self):
        draw_tree = False
        # Test with no targets
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets, draw_tree=draw_tree,
                                         expert_answers_dir=self.expert_answers_dir,
                                         expert_answers_file="grdr_expert_answers_fit_no_targets",
                                         load_answers=True, save_answers=False, append=False, no_targets=True)
        if draw_tree:
            for conclusion_name, rdr in grdr.start_rules_dict.items():
                render_tree(rdr.start_rule, use_dot_exporter=True,
                            filename=self.test_results_dir + f"/grdr_no_targets_{conclusion_name}")
        for case, case_targets in zip(self.all_cases[:20], all_targets):
            cat = grdr.classify(case)
            for cat_name, cat_val in cat.items():
                if cat_name in case_targets:
                    self.assertEqual(make_set(cat_val), make_set(case_targets[cat_name]))

    def test_write_grdr_no_targets_to_python_file(self):
        # Test with no targets
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets, draw_tree=False,
                                         expert_answers_dir=self.expert_answers_dir,
                                         expert_answers_file="grdr_expert_answers_fit_no_targets",
                                         load_answers=True, save_answers=False, append=False, no_targets=True)
        model_dir = self.generated_rdrs_dir + '/grdr_no_targets'
        grdr._write_to_python(model_dir)
        classify_species_grdr = grdr.get_rdr_classifier_from_python_file(model_dir)
        for case, case_targets in zip(self.all_cases[:20], all_targets):
            cat = classify_species_grdr(case)
            for cat_name, cat_val in cat.items():
                if cat_name in case_targets:
                    self.assertEqual(make_set(cat_val), make_set(case_targets[cat_name]))

    def test_fit_multi_line_scrdr(self):
        n = 20
        scrdr, _ = get_fit_scrdr(self.all_cases[:n], self.targets[:n], draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_multi_line_expert_answers_fit",
                                 load_answers=True,
                                 save_answers=False)
        # render_tree(scrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + f"/scrdr_multi_line")

    def test_write_multi_line_scrdr_to_python_file(self):
        n = 20
        scrdr, _ = get_fit_scrdr(self.all_cases[:n], self.targets[:n], draw_tree=False,
                                 expert_answers_dir=self.expert_answers_dir,
                                 expert_answers_file="scrdr_multi_line_expert_answers_fit",
                                 load_answers=True)
        model_dir = self.generated_rdrs_dir + '/scrdr_multi_line'
        scrdr._write_to_python(model_dir)
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(model_dir)
        for case, target in zip(self.all_cases[:n], self.targets[:n]):
            cat = classify_species_scrdr(case)
            self.assertEqual(cat, target)

    def test_write_scrdr_to_python_file(self):
        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets)
        model_dir = self.generated_rdrs_dir + '/scrdr'
        scrdr._write_to_python(model_dir)
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(model_dir)
        for case, target in zip(self.all_cases, self.targets):
            cat = classify_species_scrdr(case)
            self.assertEqual(cat, target)

    def test_update_rdr_from_python_file(self):
        scrdr, _ = get_fit_scrdr(self.all_cases, self.targets)
        modified_model_dir = self.generated_rdrs_dir + '/scrdr_modified'
        scrdr._write_to_python(modified_model_dir)
        filepath = os.path.join(modified_model_dir, f"{scrdr.generated_python_defs_file_name}.py")
        func_name = f"conditions_{scrdr.start_rule.uid}"
        first_rule_conditions, line_numbers = extract_function_source(filepath,
                                                                      func_name,
                                                                      join_lines=False,
                                                                      return_line_numbers=True)
        self.assertEqual(first_rule_conditions[func_name][-1], "    return case.milk == 1")
        # modify the condition to be case.milk==0
        with open(filepath, "r") as f:
            lines = f.readlines()
        lines[line_numbers[-1][-1]-1] = "    return case.milk == 0\n"
        with open(filepath, "w") as f:
            f.writelines(lines)
        first_rule_conditions, line_numbers = extract_function_source(filepath,
                                                                      func_name,
                                                                      join_lines=False,
                                                                      return_line_numbers=True)
        self.assertEqual(first_rule_conditions[func_name][-1], "    return case.milk == 0")
        scrdr: RDRWithCodeWriter
        scrdr.update_from_python(modified_model_dir)
        self.assertEqual(scrdr.start_rule.conditions.user_input.strip().split('\n')[-1].strip(),
                         "return case.milk == 0")
        classify_species_scrdr = scrdr.get_rdr_classifier_from_python_file(modified_model_dir)
        for case, target in zip(self.all_cases, self.targets):
            if case.milk == 0:
                cat = classify_species_scrdr(case)
                self.assertEqual(cat, Species.mammal)

    def test_write_mcrdr_to_python_file(self):
        mcrdr = get_fit_mcrdr(self.all_cases, self.targets)
        model_dir = self.generated_rdrs_dir + '/mcrdr'
        mcrdr._write_to_python(model_dir)
        classify_species_mcrdr = mcrdr.get_rdr_classifier_from_python_file(model_dir)
        for case, target in zip(self.all_cases, self.targets):
            cat = classify_species_mcrdr(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_write_mcrdr_multi_line_to_python_file(self):
        n = 20
        mcrdr = get_fit_mcrdr(self.all_cases[:n], self.targets[:n], draw_tree=False,
                              expert_answers_dir=self.expert_answers_dir,
                              expert_answers_file="mcrdr_multi_line_expert_answers_fit",
                              load_answers=True,
                              save_answers=False)
        model_dir = self.generated_rdrs_dir + '/mcrdr_multi_line'
        mcrdr._write_to_python(model_dir)
        classify_species_mcrdr = mcrdr.get_rdr_classifier_from_python_file(model_dir)
        for case, target in zip(self.all_cases[:n], self.targets[:n]):
            cat = classify_species_mcrdr(case)
            self.assertEqual(make_set(cat), make_set(target))

    def test_write_grdr_to_python_file(self):
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets)
        model_dir = self.generated_rdrs_dir + '/grdr'
        grdr._write_to_python(model_dir)
        classify_animal_grdr = grdr.get_rdr_classifier_from_python_file(model_dir)
        for case, case_targets in zip(self.all_cases[:len(all_targets)], all_targets):
            cat = classify_animal_grdr(case)
            for cat_name, cat_val in cat.items():
                self.assertEqual(make_set(cat_val), make_set(case_targets[cat_name]))

    def test_classify_mcrdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/mcrdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        mcrdr = MultiClassRDR()
        cats = mcrdr.fit_case(self.case_queries[0],
                              expert=expert)

        self.assertEqual(cats[0], self.targets[0])

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_only(self):
        use_loaded_answers = True
        draw_tree = False
        save_answers = False
        filename = self.expert_answers_dir + "/mcrdr_expert_answers_stop_only_fit"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR()
        case_queries = self.case_queries
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        # render_tree(mcrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + f"/mcrdr_stop_only")
        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            self.assertEqual(make_set(cat), make_set(case_query.target_value))
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_plus_rule(self):
        use_loaded_answers = True
        draw_tree = False
        save_answers = False
        append = False
        filename = self.expert_answers_dir + "/mcrdr_stop_plus_rule_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers, append=append)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR(mode=MCRDRMode.StopPlusRule)
        case_queries = self.case_queries
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        # render_tree(mcrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + f"/mcrdr_stop_plus_rule")
        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            self.assertEqual(make_set(cat), make_set(case_query.target_value))
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_mcrdr_stop_plus_rule_combined(self):
        use_loaded_answers = True
        save_answers = False
        draw_tree = False
        append = False
        filename = self.expert_answers_dir + "/mcrdr_stop_plus_rule_combined_expert_answers_fit"
        expert = Human(use_loaded_answers=use_loaded_answers, append=append)
        if use_loaded_answers:
            expert.load_answers(filename)
        mcrdr = MultiClassRDR(mode=MCRDRMode.StopPlusRuleCombined)
        case_queries = self.case_queries
        mcrdr.fit(case_queries, expert=expert, animate_tree=draw_tree)
        # render_tree(mcrdr.start_rule, use_dot_exporter=True,
        #             filename=self.test_results_dir + f"/mcrdr_stop_plus_rule_combined")
        for case_query in case_queries:
            cat = mcrdr.classify(case_query.case)
            self.assertEqual(make_set(cat), make_set(case_query.target_value))
        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_classify_grdr(self):
        use_loaded_answers = True
        save_answers = False
        filename = self.expert_answers_dir + "/grdr_expert_answers_classify"
        expert = Human(use_loaded_answers=use_loaded_answers)
        if use_loaded_answers:
            expert.load_answers(filename)

        grdr = GeneralRDR()

        targets = [self.targets[0], Habitat.land]
        attribute_names = [t.__class__.__name__.lower() for t in targets]
        targets = dict(zip(attribute_names, targets))
        case_queries = [CaseQuery(self.all_cases[0], a, (type(t),), True if a == 'species' else False,
                                  _target=t) for a, t in targets.items()]
        grdr.fit(case_queries, expert=expert)
        cats = grdr.classify(self.all_cases[0])
        for cat_name, value in cats.items():
            self.assertEqual(make_set(value), make_set(targets[cat_name]))

        if save_answers:
            cwd = os.getcwd()
            file = os.path.join(cwd, filename)
            expert.save_answers(file)

    def test_fit_grdr(self):
        grdr, all_targets = get_fit_grdr(self.all_cases, self.targets, draw_tree=False,
                                         load_answers=True, save_answers=False)
        # for conclusion_name, rdr in grdr.start_rules_dict.items():
        #     render_tree(rdr.start_rule, use_dot_exporter=True,
        #                 filename=self.test_results_dir + f"/grdr_{conclusion_name}")




import os.path
import unittest

try:
    from PyQt6.QtWidgets import QApplication
    from ripple_down_rules.user_interface.gui import RDRCaseViewer, style
except ImportError as e:
    print(f"Skipping GUI tests due to missing PyQt6: {e}")
    QApplication = None
    RDRCaseViewer = None
    style = None
from typing_extensions import List

from datasets import load_zoo_dataset, Species
from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from test_helpers.helpers import get_fit_grdr
from test_object_diagram import Person, Address


@unittest.skipIf(QApplication is None, "GUI tests need PyQt6,"
                                       "and they need visual inspection by a user and cannot be run automatically.")
class GUITestCase(unittest.TestCase):
    """Test case for the GUI components of the ripple down rules package."""
    app: QApplication
    viewer: RDRCaseViewer
    cq: CaseQuery
    cases: List[Case]
    person: Person

    @classmethod
    def setUpClass(cls):
        print("Setting up GUI test case...")
        cls.app = QApplication([])
        cls.cases, cls.targets = load_zoo_dataset(cache_file=f"{os.path.dirname(__file__)}/../test_results/zoo")
        cls.cq = CaseQuery(cls.cases[0], "species", (Species,), True, _target=cls.targets[0])
        cls.viewer = RDRCaseViewer(save_file=f"{os.path.dirname(__file__)}/../test_results/grdr_viewer")
        cls.person = Person("Ahmed", Address("Cairo"))

    def test_change_title_text(self):
        self.viewer.show()
        self.app.exec()
        self.viewer.title_label.setText(style("Changed Title", "o", 28, 'bold'))
        self.viewer.show()
        self.app.exec()

    def test_update_image(self):
        self.viewer.obj_diagram_viewer.update_image(f"{os.path.dirname(__file__)}/../test_helpers/object_diagram_case_query.png")
        self.viewer.show()
        self.app.exec()
        self.viewer.obj_diagram_viewer.update_image(f"{os.path.dirname(__file__)}/../test_helpers/object_diagram_person.png")
        self.viewer.show()
        self.app.exec()

    def test_update_for_obj(self):
        self.viewer.update_for_object(self.cq, "CaseQuery")
        self.viewer.show()
        self.app.exec()
        self.viewer.update_for_object(self.person, "Person")
        self.viewer.show()
        self.app.exec()

    def test_save_button(self):
        grdr, _ = get_fit_grdr(self.cases, self.targets)
        grdr.set_viewer(self.viewer)
        self.viewer.title_label.setText(style("Press `Save` To Test", "o", 28, 'bold'))
        self.viewer.show()
        self.app.exec()
        grdr_viewer_dir = f"{os.path.dirname(__file__)}/../test_results/grdr_viewer"
        self.assertTrue(os.path.exists(f"{grdr_viewer_dir}/{grdr.model_name}/rdr_metadata/{grdr.model_name}.json"))




from __future__ import annotations

import ast
import builtins
import copyreg
import importlib
import json
import logging
import os
import re
import threading
import uuid
from collections import UserDict
from copy import deepcopy, copy
from dataclasses import is_dataclass, fields
from enum import Enum
from textwrap import dedent
from types import NoneType
from typing import List

from casadi import conic_n_out

try:
    import matplotlib
    from matplotlib import pyplot as plt
    Figure = plt.Figure
except ImportError as e:
    matplotlib = None
    plt = None
    Figure = None
    logging.debug(f"{e}: matplotlib is not installed")

try:
    import networkx as nx
except ImportError as e:
    nx = None
    logging.debug(f"{e}: networkx is not installed")

import requests
from anytree import Node, RenderTree
from anytree.exporter import DotExporter
from sqlalchemy import MetaData, inspect
from sqlalchemy.orm import Mapped, registry, class_mapper, DeclarativeBase as SQLTable, Session
from tabulate import tabulate
from typing_extensions import Callable, Set, Any, Type, Dict, TYPE_CHECKING, get_type_hints, \
    get_origin, get_args, Tuple, Optional, List, Union, Self


if TYPE_CHECKING:
    from .datastructures.case import Case
    from .datastructures.dataclasses import CaseQuery

import ast


def str_to_snake_case(snake_str: str) -> str:
    """
    Convert a string to snake case.

    :param snake_str: The string to convert.
    :return: The converted string.
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', snake_str)
    s1 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
    # remove redundant underscores
    s1 = re.sub(r'_{2,}', '_', s1)
    # remove leading and trailing underscores
    s1 = re.sub(r'^_|_$', '', s1)
    return s1


def are_results_subclass_of_types(result_types: List[Any], types_: List[Type]) -> bool:
    """
    Check if all results are subclasses of the given types.

    :param result_types: The list of result types to check.
    :param types_: The list of types to check against.
    :return: True if all results are subclasses of the given types, False otherwise.
    """
    for rt in result_types:
        if not any(issubclass(rt, t) for t in types_):
            return False
    return True


def get_imports_from_types(types: List[Type]) -> List[str]:
    """
    Get the import statements for a list of types.

    :param types: The types to get the import statements for.
    :return: The import statements as a string.
    """
    imports = map(get_import_from_type, types)
    return list({i for i in imports if i is not None})


def get_import_from_type(type_: Type) -> Optional[str]:
    """
    Get the import statement for a given type.

    :param type_: The type to get the import statement for.
    :return: The import statement as a string.
    """
    if hasattr(type_, "__module__") and hasattr(type_, "__name__"):
        if type_.__module__ == "builtins":
            return
        return f"from {type_.__module__} import {type_.__name__}"


def get_imports_from_scope(scope: Dict[str, Any]) -> List[str]:
    """
    Get the imports from the given scope.

    :param scope: The scope to get the imports from.
    :return: The imports as a string.
    """
    imports = []
    for k, v in scope.items():
        if not hasattr(v, "__module__") or not hasattr(v, "__name__") or v.__module__ is None:
            continue
        imports.append(f"from {v.__module__} import {v.__name__}")
    return imports


def extract_imports(file_path: Optional[str] = None, tree: Optional[ast.AST] = None) -> Dict[str, Any]:
    if tree is None:
        if file_path is None:
            raise ValueError("Either file_path or tree must be provided")
        with open(file_path, "r") as f:
            tree = ast.parse(f.read(), filename=file_path)

    scope = {}

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                module_name = alias.name
                asname = alias.asname or alias.name
                try:
                    scope[asname] = importlib.import_module(module_name)
                except ImportError as e:
                    print(f"Could not import {module_name}: {e}")
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            for alias in node.names:
                name = alias.name
                asname = alias.asname or name
                try:
                    module = importlib.import_module(module_name)
                    scope[asname] = getattr(module, name)
                except (ImportError, AttributeError) as e:
                    logging.warning(f"Could not import {module_name}: {e} while extracting imports from {file_path}")

    return scope


def extract_function_source(file_path: str,
                            function_names: List[str], join_lines: bool = True,
                            return_line_numbers: bool = False,
                            include_signature: bool = True) \
        -> Union[Dict[str, Union[str, List[str]]],
        Tuple[Dict[str, Union[str, List[str]]], List[Tuple[int, int]]]]:
    """
    Extract the source code of a function from a file.

    :param file_path: The path to the file.
    :param function_names: The names of the functions to extract.
    :param join_lines: Whether to join the lines of the function.
    :param return_line_numbers: Whether to return the line numbers of the function.
    :param include_signature: Whether to include the function signature in the source code.
    :return: A dictionary mapping function names to their source code as a string if join_lines is True,
     otherwise as a list of strings.
    """
    with open(file_path, "r") as f:
        source = f.read()

    # Parse the source code into an AST
    tree = ast.parse(source)
    function_names = make_list(function_names)
    functions_source: Dict[str, Union[str, List[str]]] = {}
    line_numbers = []
    for node in tree.body:
        if isinstance(node, ast.FunctionDef) and (node.name in function_names or len(function_names) == 0):
            # Get the line numbers of the function
            lines = source.splitlines()
            func_lines = lines[node.lineno - 1:node.end_lineno]
            if not include_signature:
                func_lines = func_lines[1:]
            line_numbers.append((node.lineno, node.end_lineno))
            functions_source[node.name] = dedent("\n".join(func_lines)) if join_lines else func_lines
            if (len(functions_source) >= len(function_names)) and (not len(function_names) == 0):
                break
    if len(functions_source) < len(function_names):
        raise ValueError(f"Could not find all functions in {file_path}: {function_names} not found,"
                         f"functions not found: {set(function_names) - set(functions_source.keys())}")
    if return_line_numbers:
        return functions_source, line_numbers
    return functions_source


def encapsulate_user_input(user_input: str, func_signature: str, func_doc: Optional[str] = None) -> str:
    """
    Encapsulate the user input string with a function definition.

    :param user_input: The user input string.
    :param func_signature: The function signature to use for encapsulation.
    :param func_doc: The function docstring to use for encapsulation.
    :return: The encapsulated user input string.
    """
    if func_signature not in user_input:
        new_user_input = func_signature + "\n    "
        if func_doc is not None:
            new_user_input += f"\"\"\"{func_doc}\"\"\"" + "\n    "
        if "return " not in user_input:
            if '\n' not in user_input:
                new_user_input += f"return {user_input}"
            else:
                raise ValueError("User input must contain a return statement or be a single line.")
        else:
            for cl in user_input.split('\n'):
                sub_code_lines = cl.split('\n')
                new_user_input += '\n    '.join(sub_code_lines) + '\n    '
    else:
        new_user_input = user_input
    return new_user_input


def build_user_input_from_conclusion(conclusion: Any) -> str:
    """
    Build a user input string from the conclusion.

    :param conclusion: The conclusion to use for the callable expression.
    :return: The user input string.
    """

    # set user_input to the string representation of the conclusion
    if isinstance(conclusion, set):
        user_input = '{' + f"{', '.join([conclusion_to_str(t) for t in conclusion])}" + '}'
    elif isinstance(conclusion, list):
        user_input = '[' + f"{', '.join([conclusion_to_str(t) for t in conclusion])}" + ']'
    elif isinstance(conclusion, tuple):
        user_input = '(' + f"{', '.join([conclusion_to_str(t) for t in conclusion])}" + ')'
    else:
        user_input = conclusion_to_str(conclusion)

    return user_input


def conclusion_to_str(conclusion_: Any) -> str:
    if isinstance(conclusion_, Enum):
        return type(conclusion_).__name__ + '.' + conclusion_.name
    else:
        return str(conclusion_)


def update_case(case_query: CaseQuery, conclusions: Dict[str, Any]):
    """
    Update the case with the conclusions.

    :param case_query: The case query that contains the case to update.
    :param conclusions: The conclusions to update the case with.
    """
    if not conclusions:
        return
    if len(conclusions) == 0:
        return
    if isinstance(case_query.original_case, SQLTable) or is_dataclass(case_query.original_case):
        for conclusion_name, conclusion in conclusions.items():
            attribute = getattr(case_query.case, conclusion_name)
            if conclusion_name == case_query.attribute_name:
                attribute_type = case_query.attribute_type
            else:
                attribute_type = (get_case_attribute_type(case_query.original_case, conclusion_name, attribute),)
            if isinstance(attribute, set):
                for c in conclusion:
                    attribute.update(make_set(c))
            elif isinstance(attribute, list):
                attribute.extend(conclusion)
            elif any(at in {List, list} for at in attribute_type):
                attribute = [] if attribute is None else attribute
                attribute.extend(conclusion)
            elif any(at in {Set, set} for at in attribute_type):
                attribute = set() if attribute is None else attribute
                for c in conclusion:
                    attribute.update(make_set(c))
            elif is_iterable(conclusion) and len(conclusion) == 1 \
                    and any(at is type(list(conclusion)[0]) for at in attribute_type):
                setattr(case_query.case, conclusion_name, list(conclusion)[0])
            elif not is_iterable(conclusion) and any(at is type(conclusion) for at in attribute_type):
                setattr(case_query.case, conclusion_name, conclusion)
            else:
                raise ValueError(f"Unknown type or type mismatch for attribute {conclusion_name} with type "
                                 f"{case_query.attribute_type} with conclusion "
                                 f"{conclusion} of type {type(conclusion)}")
    else:
        case_query.case.update(conclusions)


def is_conflicting(conclusion: Any, target: Any) -> bool:
    """
    :param conclusion: The conclusion to check.
    :param target: The target to compare the conclusion with.
    :return: Whether the conclusion is conflicting with the target by have different values for same type categories.
    """
    return have_common_types(conclusion, target) and not make_set(conclusion).issubset(make_set(target))


def have_common_types(conclusion: Any, target: Any) -> bool:
    """
    :param conclusion: The conclusion to check.
    :param target: The target to compare the conclusion with.
    :return: Whether the conclusion shares some types with the target.
    """
    target_types = {type(t) for t in make_set(target)}
    conclusion_types = {type(c) for c in make_set(conclusion)}
    common_types = conclusion_types.intersection(target_types)
    return len(common_types) > 0


def calculate_precision_and_recall(pred_cat: Dict[str, Any], target: Dict[str, Any]) -> Tuple[
    List[bool], List[bool]]:
    """
    :param pred_cat: The predicted category.
    :param target: The target category.
    :return: The precision and recall of the classifier.
    """
    recall = []
    precision = []
    for pred_key, pred_value in pred_cat.items():
        if pred_key not in target:
            continue
        precision.extend([v in make_set(target[pred_key]) for v in make_set(pred_value)])
    for target_key, target_value in target.items():
        if target_key not in pred_cat:
            recall.append(False)
            continue
        recall.extend([v in make_set(pred_cat[target_key]) for v in make_set(target_value)])
    return precision, recall


def ask_llm(prompt):
    try:
        response = requests.post("http://localhost:11434/api/generate", json={
            "model": "codellama:7b-instruct",  # or "phi"
            "prompt": prompt,
            "stream": False,
        })
        result = response.json()
        return result.get("response", "").strip()
    except Exception as e:
        return f"❌ Local LLM error: {e}"


def get_case_attribute_type(original_case: Any, attribute_name: str,
                            known_value: Optional[Any] = None) -> Type:
    """
    :param original_case: The case to get the attribute from.
    :param attribute_name: The name of the attribute.
    :param known_value: A known value of the attribute.
    :return: The type of the attribute.
    """
    if known_value is not None:
        return type(known_value)
    elif hasattr(original_case, attribute_name):
        hint, origin, args = get_hint_for_attribute(attribute_name, original_case)
        if origin is not None:
            origin = typing_to_python_type(origin)
        if origin == Union:
            if len(args) == 2:
                if args[1] is type(None):
                    return typing_to_python_type(args[0])
                elif args[0] is type(None):
                    return typing_to_python_type(args[1])
            elif len(args) == 1:
                return typing_to_python_type(args[0])
            else:
                raise ValueError(f"Union with more than 2 types is not supported: {args}")
        elif origin is not None:
            return origin
        if hint is not None:
            return typing_to_python_type(hint)


def conclusion_to_json(conclusion):
    if is_iterable(conclusion):
        conclusions = {'_type': get_full_class_name(type(conclusion)), 'value': []}
        for c in conclusion:
            conclusions['value'].append(conclusion_to_json(c))
    elif hasattr(conclusion, 'to_json'):
        conclusions = conclusion.to_json()
    else:
        conclusions = {'_type': get_full_class_name(type(conclusion)), 'value': conclusion}
    return conclusions


def contains_return_statement(source: str) -> bool:
    """
    :param source: The source code to check.
    :return: True if the source code contains a return statement, False otherwise.
    """
    try:
        tree = ast.parse(source)
        for node in tree.body:
            if isinstance(node, ast.Return):
                return True
        return False
    except SyntaxError:
        return False


def get_names_used(node):
    return {n.id for n in ast.walk(node) if isinstance(n, ast.Name)}


def extract_dependencies(code_lines):
    full_code = '\n'.join(code_lines)
    tree = ast.parse(full_code)
    final_stmt = tree.body[-1]

    if not isinstance(final_stmt, ast.Return):
        raise ValueError("Last line is not a return statement")
    if final_stmt.value is None:
        raise ValueError("Return statement has no value")
    needed = get_names_used(final_stmt.value)
    required_lines = []
    line_map = {id(node): i for i, node in enumerate(tree.body)}

    def handle_stmt(stmt, needed):
        keep = False
        if isinstance(stmt, ast.Assign):
            targets = [t.id for t in stmt.targets if isinstance(t, ast.Name)]
            if any(t in needed for t in targets):
                needed.update(get_names_used(stmt.value))
                keep = True
        elif isinstance(stmt, ast.AugAssign):
            if isinstance(stmt.target, ast.Name) and stmt.target.id in needed:
                needed.update(get_names_used(stmt.value))
                keep = True
        elif isinstance(stmt, ast.FunctionDef):
            if stmt.name in needed:
                for n in ast.walk(stmt):
                    if isinstance(n, ast.Name):
                        needed.add(n.id)
                keep = True
        elif isinstance(stmt, (ast.For, ast.While, ast.If)):
            # Check if any of the body statements interact with needed variables
            for substmt in stmt.body + getattr(stmt, 'orelse', []):
                if handle_stmt(substmt, needed):
                    keep = True
            # Also check the condition (test or iter)
            if isinstance(stmt, ast.For):
                if isinstance(stmt.target, ast.Name) and stmt.target.id in needed:
                    keep = True
                needed.update(get_names_used(stmt.iter))
            elif isinstance(stmt, ast.If) or isinstance(stmt, ast.While):
                needed.update(get_names_used(stmt.test))

        return keep

    for stmt in reversed(tree.body[:-1]):
        if handle_stmt(stmt, needed):
            # check if the statement is a function, if so then all its lines not just the line in line_map are needed.
            if isinstance(stmt, ast.FunctionDef):
                start_code_line = line_map[id(stmt)]
                end_code_line = start_code_line + stmt.end_lineno
                required_lines.extend(code_lines[start_code_line:end_code_line])
            else:
                required_lines.insert(0, code_lines[line_map[id(stmt)]])

    required_lines.append(code_lines[-1])  # Always include return
    return required_lines


def serialize_dataclass(obj: Any, seen=None) -> Any:
    """
    Recursively serialize a dataclass to a dictionary. If the dataclass contains any nested dataclasses, they will be
    serialized as well. If the object is not a dataclass, it will be returned as is.

    :param obj: The dataclass to serialize.
    :return: The serialized dataclass as a dictionary or the object itself if it is not a dataclass.
    """
    if seen is None:
        seen = {}

    obj_id = id(obj)
    if obj_id in seen:
        return {'$ref': seen[obj_id]}

    if is_dataclass(obj):
        uid = str(uuid.uuid4())
        seen[obj_id] = uid
        result = {
            '$id': uid,
            "__dataclass__": f"{obj.__class__.__module__}.{obj.__class__.__qualname__}",
            'fields': {}
        }
        for f in fields(obj):
            value = getattr(obj, f.name)
            result['fields'][f.name] = serialize_dataclass(value, seen)
        return result
    elif isinstance(obj, list):
        return [serialize_dataclass(v, seen) for v in obj]
    elif isinstance(obj, dict):
        serialized_dict = {}
        for k, v in obj.items():
            if not isinstance(k, (str, int, bool, float, type(None))):
                continue
            serialized_dict[k] = serialize_dataclass(v, seen)
        return serialized_dict
    else:
        try:
            json.dumps(obj)  # Check if the object is JSON serializable
            return obj
        except TypeError:
            return None


def deserialize_dataclass(data: Any, refs: Optional[Dict[str, Any]] = None) -> Any:
    refs = {} if refs is None else refs
    preloaded = preload_serialized_objects(data, refs)
    return resolve_refs(preloaded, refs)


def preload_serialized_objects(data: Any, refs: Dict[str, Any] = None) -> Any:
    """
    Recursively deserialize a dataclass from a dictionary, if the dictionary contains a key "__dataclass__" (Most likely
    created by the serialize_dataclass function), it will be treated as a dataclass and deserialized accordingly,
    otherwise it will be returned as is.

    :param data: The dictionary to deserialize.
    :return: The deserialized dataclass.
    """
    if refs is None:
        refs = {}

    if isinstance(data, dict):

        if '$ref' in data:
            ref_id = data['$ref']
            if ref_id not in refs:
                return {'$ref': data['$ref']}
            return refs[ref_id]

        elif '$id' in data and '__dataclass__' in data and 'fields' in data:
            cls_path = data['__dataclass__']
            module_name, class_name = cls_path.rsplit('.', 1)
            cls = getattr(importlib.import_module(module_name), class_name)

            dummy_instance = cls.__new__(cls)  # Don't call __init__ yet
            refs[data['$id']] = dummy_instance

            for f in fields(cls):
                raw_value = data['fields'].get(f.name)
                value = preload_serialized_objects(raw_value, refs)
                setattr(dummy_instance, f.name, value)

            return dummy_instance

        else:
            return {k: preload_serialized_objects(v, refs) for k, v in data.items()}

    elif isinstance(data, list):
        return [preload_serialized_objects(item, refs) for item in data]
    elif isinstance(data, dict):
        return {k: preload_serialized_objects(v, refs) for k, v in data.items()}

    return data  # Primitive


def resolve_refs(obj, refs, seen=None):
    if seen is None:
        seen = {}

    obj_id = id(obj)
    if obj_id in seen:
        return seen[obj_id]

    # Resolve if dict with $ref
    if isinstance(obj, dict) and '$ref' in obj:
        ref_id = obj['$ref']
        if ref_id not in refs:
            raise KeyError(f"$ref to unknown ID: {ref_id}")
        return refs[ref_id]

    elif is_dataclass(obj):
        seen[obj_id] = obj  # Mark before diving deeper
        for f in fields(obj):
            val = getattr(obj, f.name)
            resolved = resolve_refs(val, refs, seen)
            setattr(obj, f.name, resolved)
        return obj

    elif isinstance(obj, list):
        resolved_list = []
        seen[obj_id] = resolved_list
        for item in obj:
            resolved_list.append(resolve_refs(item, refs, seen))
        return resolved_list

    elif isinstance(obj, dict):
        resolved_dict = {}
        seen[obj_id] = resolved_dict
        for k, v in obj.items():
            resolved_dict[k] = resolve_refs(v, refs, seen)
        return resolved_dict

    return obj  # Primitive


def typing_to_python_type(typing_hint: Type) -> Type:
    """
    Convert a typing hint to a python type.

    :param typing_hint: The typing hint to convert.
    :return: The python type.
    """
    if typing_hint in [list, List]:
        return list
    elif typing_hint in [tuple, Tuple]:
        return tuple
    elif typing_hint in [set, Set]:
        return set
    elif typing_hint in [dict, Dict]:
        return dict
    else:
        return typing_hint


def capture_variable_assignment(code: str, variable_name: str) -> Optional[str]:
    """
    Capture the assignment of a variable in the given code.

    :param code: The code to analyze.
    :param variable_name: The name of the variable to capture.
    :return: The assignment statement or None if not found.
    """
    tree = ast.parse(code)
    assignment = None
    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == variable_name:
                    # now extract the right side of the assignment
                    assignment = ast.get_source_segment(code, node.value)
                    break
        if assignment is not None:
            break
    return assignment


def get_func_rdr_model_path(func: Callable, model_dir: str) -> str:
    """
    :param func: The function to get the model path for.
    :param model_dir: The directory to save the model to.
    :return: The path to the model file.
    """
    return os.path.join(model_dir, f"{get_func_rdr_model_name(func)}.json")


def get_func_rdr_model_name(func: Callable, include_file_name: bool = False) -> str:
    """
    :param func: The function to get the model name for.
    :return: The name of the model.
    """
    func_name = get_method_name(func)
    func_class_name = get_method_class_name_if_exists(func)
    if include_file_name:
        func_file_name = get_method_file_name(func).split(os.sep)[-1].split('.')[0]
        model_name = func_file_name + '_'
    else:
        model_name = ''
    model_name += f"{func_class_name}_" if func_class_name else ""
    model_name += f"{func_name}"
    return model_name


def extract_bracket_arguments(val: str) -> List[str]:
    """
    Extract arguments inside brackets into a list.

    :param val: The string containing brackets.
    :return: List of arguments inside brackets.
    """
    if '[' not in val:
        return [val]
    args_start = val.find('[')
    args_end = val.rfind(']')
    if args_end == -1:
        return [val]
    base_type = val[:args_start]
    args = val[args_start + 1:args_end].split(',')
    args = [arg.strip() for arg in args]
    return [base_type] + args


def typing_hint_to_str(type_hint: Any) -> Tuple[str, List[str]]:
    """
    Convert a typing hint to a string.

    :param type_hint: The typing hint to convert.
    :return: The string representation of the typing hint.
    """
    val = (str(type_hint).strip("<>")
           .replace("class ", "")
           # .replace("typing.", "")
           .replace("'", ""))
    all_args = []
    if '[' in val:
        args = extract_bracket_arguments(val)
        args_with_brackets = [arg for arg in args if '[' in arg]
        all_args.extend([arg for arg in args if '[' not in arg])
        while args_with_brackets:
            for arg in args:
                if '[' in arg:
                    sub_args = extract_bracket_arguments(arg)
                    args_with_brackets.remove(arg)
                    all_args.extend([sarg for sarg in sub_args if '[' not in sarg])
                    args_with_brackets.extend([sarg for sarg in sub_args if '[' in sarg])
                elif arg not in all_args:
                    all_args.append(arg)
            args = args_with_brackets
        for arg in all_args:
            val = val.replace(arg, arg.split('.')[-1])
    else:
        val = val.split('.')[-1]
    return val, all_args


def get_method_args_as_dict(method: Callable, *args, **kwargs) -> Dict[str, Any]:
    """
    Get the arguments of a method as a dictionary.

    :param method: The method to get the arguments from.
    :param args: The positional arguments.
    :param kwargs: The keyword arguments.
    :return: A dictionary of the arguments.
    """
    func_arg_names = method.__code__.co_varnames
    func_arg_names = list(map(lambda arg_name: f"{arg_name}_" if arg_name in ["self", "cls"] else arg_name,
                              func_arg_names))
    func_arg_values = args + tuple(kwargs.values())
    return dict(zip(func_arg_names, func_arg_values))


def get_method_name(method: Callable) -> str:
    """
    Get the name of a method.

    :param method: The method to get the name of.
    :return: The name of the method.
    """
    return method.__name__ if hasattr(method, "__name__") else str(method)


def get_method_class_name_if_exists(method: Callable) -> Optional[str]:
    """
    Get the class name of a method if it has one.

    :param method: The method to get the class name of.
    :return: The class name of the method.
    """
    if hasattr(method, "__self__"):
        if hasattr(method.__self__, "__class__"):
            return method.__self__.__class__.__name__
    return method.__qualname__.split('.')[0] if hasattr(method, "__qualname__") else None


def get_method_class_if_exists(method: Callable, *args) -> Optional[Type]:
    """
    Get the class of a method if it has one.

    :param method: The method to get the class of.
    :return: The class of the method, if it exists otherwise None.
    """
    if hasattr(method, "__self__"):
        if hasattr(method.__self__, "__class__"):
            return method.__self__.__class__
    elif method.__code__.co_varnames:
        if method.__code__.co_varnames[0] == 'self':
            return args[0].__class__
        elif method.__code__.co_varnames[0] == 'cls':
            return args[0]
    return None


def get_method_file_name(method: Callable) -> str:
    """
    Get the file name of a method.

    :param method: The method to get the file name of.
    :return: The file name of the method.
    """
    return method.__code__.co_filename


def flatten_list(a: List):
    a_flattened = []
    for c in a:
        if is_iterable(c):
            a_flattened.extend(list(c))
        else:
            a_flattened.append(c)
    return a_flattened


def make_list(value: Any) -> List:
    """
    Make a list from a value.

    :param value: The value to make a list from.
    """
    return list(value) if is_iterable(value) else [value]


def is_iterable(obj: Any) -> bool:
    """
    Check if an object is iterable.

    :param obj: The object to check.
    """
    return hasattr(obj, "__iter__") and not isinstance(obj, (str, type, bytes, bytearray))


def get_type_from_string(type_path: str):
    """
    Get a type from a string describing its path using the format "module_path.ClassName".

    :param type_path: The path to the type.
    """
    module_path, class_name = type_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    if module == builtins and class_name == 'NoneType':
        return type(None)
    return getattr(module, class_name)


def get_full_class_name(cls):
    """
    Returns the full name of a class, including the module name.
    Copied from: https://github.com/tomsch420/random-events/blob/master/src/random_events/utils.py#L6C1-L21C101

    :param cls: The class.
    :return: The full name of the class
    """
    return cls.__module__ + "." + cls.__name__


def recursive_subclasses(cls):
    """
    Copied from: https://github.com/tomsch420/random-events/blob/master/src/random_events/utils.py#L6C1-L21C101
    :param cls: The class.
    :return: A list of the classes subclasses.
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in recursive_subclasses(s)]


class SubclassJSONSerializer:
    """
    Originally from: https://github.com/tomsch420/random-events/blob/master/src/random_events/utils.py#L6C1-L21C101
    Class for automatic (de)serialization of subclasses.
    Classes that inherit from this class can be serialized and deserialized automatically by calling this classes
    'from_json' method.
    """
    data_class_refs = {}

    def to_json_file(self, filename: str):
        """
        Save the object to a json file.
        """
        data = self.to_json()
        # save the json to a file
        if not filename.endswith(".json"):
            filename += ".json"
        with open(filename, "w") as f:
            json.dump(data, f, indent=4)
        return data

    @staticmethod
    def to_json_static(obj) -> Dict[str, Any]:
        if is_dataclass(obj):
            return serialize_dataclass(obj)
        return {"_type": get_full_class_name(obj.__class__), **obj._to_json()}

    def to_json(self) -> Dict[str, Any]:
        return self.to_json_static(self)

    def _to_json(self) -> Dict[str, Any]:
        """
        Create a json dict from the object.
        """
        raise NotImplementedError()

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create a variable from a json dict.
        This method is called from the from_json method after the correct subclass is determined and should be
        overwritten by the respective subclass.

        :param data: The json dict
        :return: The deserialized object
        """
        raise NotImplementedError()

    @classmethod
    def from_json_file(cls, filename: str) -> Any:
        """
        Create an instance of the subclass from the data in the given json file.

        :param filename: The filename of the json file.
        """
        if not filename.endswith(".json"):
            filename += ".json"
        with open(filename, "r") as f:
            rdr_json = json.load(f)
        deserialized_obj = cls.from_json(rdr_json)
        cls.data_class_refs.clear()
        return deserialized_obj

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create the correct instanceof the subclass from a json dict.

        :param data: The json dict
        :return: The correct instance of the subclass
        """
        if data is None:
            return None
        if isinstance(data, list):
            # if the data is a list, deserialize it
            return [cls.from_json(d) for d in data]
        elif isinstance(data, dict):
            if '__dataclass__' in data:
                # if the data is a dataclass, deserialize it
                return deserialize_dataclass(data, cls.data_class_refs)
            elif '_type' not in data:
                return {k: cls.from_json(v) for k, v in data.items()}
        elif not isinstance(data, dict):
            return data

        # check if type module is builtins
        data_type = get_type_from_string(data["_type"])
        if len(data) == 1:
            return data_type
        if data_type == NoneType:
            return None
        if data_type.__module__ == 'builtins':
            if is_iterable(data['value']) and not isinstance(data['value'], dict):
                return data_type([cls.from_json(d) for d in data['value']])
            return data_type(data["value"])
        if get_full_class_name(cls) == data["_type"]:
            data.pop("_type")
            return cls._from_json(data)
        for subclass in recursive_subclasses(SubclassJSONSerializer):
            if get_full_class_name(subclass) == data["_type"]:
                subclass_data = deepcopy(data)
                subclass_data.pop("_type")
                return subclass._from_json(subclass_data)

        raise ValueError("Unknown type {}".format(data["_type"]))


def _pickle_thread(thread_obj) -> Any:
    """Return a plain object with user-defined attributes but no thread behavior."""

    class DummyThread:
        pass

    dummy = DummyThread()
    # Copy only non-thread-related attributes
    for attr, value in thread_obj.__dict__.items():
        print(attr)
        if not attr.startswith("_"):  # Skip internal Thread attributes
            setattr(dummy, attr, value)
    return dummy


copyreg.pickle(threading.Thread, _pickle_thread)


def copy_case(case: Union[Case, SQLTable]) -> Union[Case, SQLTable, Any]:
    """
    Copy a case.

    :param case: The case to copy.
    :return: The copied case.
    """
    if isinstance(case, SQLTable):
        return copy_orm_instance_with_relationships(case)
    else:
        # copy the case recursively for 1 level
        # try:
        #     case_copy = deepcopy(case)
        # except Exception as e:
        case_copy = copy(case)
        for attr in dir(case):
            if attr.startswith("_") or callable(getattr(case, attr)):
                continue
            attr_value = getattr(case, attr)
            if is_iterable(attr_value):
                try:
                    setattr(case_copy, attr, copy(attr_value))
                except AttributeError as e:
                    # if the attribute is not settable, just skip it
                    pass
        return case_copy


def copy_orm_instance(instance: SQLTable) -> SQLTable:
    """
    Copy an ORM instance by expunging it from the session then deep copying it and adding it back to the session. This
    is useful when you want to copy an instance and make changes to it without affecting the original instance.

    :param instance: The instance to copy.
    :return: The copied instance.
    """
    session: Session = inspect(instance).session
    if session is not None:
        session.expunge(instance)
        new_instance = deepcopy(instance)
        session.add(instance)
    else:
        new_instance = instance
    return new_instance


def copy_orm_instance_with_relationships(instance: SQLTable) -> SQLTable:
    """
    Copy an ORM instance with its relationships (i.e. its foreign keys).

    :param instance: The instance to copy.
    :return: The copied instance.
    """
    instance_cp = copy_orm_instance(instance)
    for rel in class_mapper(instance.__class__).relationships:
        related_obj = getattr(instance, rel.key)
        if related_obj is not None:
            setattr(instance_cp, rel.key, related_obj)
    return instance_cp


def get_value_type_from_type_hint(attr_name: str, obj: Any) -> Type:
    """
    Get the value type from the type hint of an object attribute.

    :param attr_name: The name of the attribute.
    :param obj: The object to get the attributes from.
    """
    hint, origin, args = get_hint_for_attribute(attr_name, obj)
    if not origin and not hint:
        if hasattr(obj, attr_name):
            attr_value = getattr(obj, attr_name)
            if attr_value is not None:
                return type(attr_value)
        raise ValueError(f"Couldn't get type for Attribute {attr_name}, please provide a type hint")
    if origin in [list, set, tuple, type, dict]:
        attr_value_type = args[0]
    elif hint:
        attr_value_type = hint
    else:
        raise ValueError(f"Attribute {attr_name} has unsupported type {hint}.")
    return attr_value_type


def get_hint_for_attribute(attr_name: str, obj: Any) -> Tuple[Optional[Type], Optional[Type], Tuple[Type]]:
    """
    Get the type hint for an attribute of an object.

    :param attr_name: The name of the attribute.
    :param obj: The object to get the attribute from.
    :return: The type hint of the attribute.
    """
    if attr_name is None or not hasattr(obj.__class__, attr_name):
        return None, None, ()
    class_attr = getattr(obj.__class__, attr_name)
    if isinstance(class_attr, property):
        if not class_attr.fget:
            raise ValueError(f"Attribute {attr_name} has no getter.")
        hint = get_type_hints(class_attr.fget)['return']
    else:
        try:
            hint = get_type_hints(obj.__class__)[attr_name]
        except KeyError:
            hint = type(class_attr)
    origin, args = get_origin_and_args_from_type_hint(hint)
    return origin, origin, args


def get_origin_and_args_from_type_hint(type_hint: Type) -> Tuple[Optional[Type], Tuple[Type]]:
    """
    Get the origin and arguments from a type hint.W

    :param type_hint: The type hint to get the origin and arguments from.
    :return: The origin and arguments of the type hint.
    """
    origin = get_origin(type_hint)
    args = get_args(type_hint)
    if origin is Mapped:
        return get_origin(args[0]), get_args(args[0])
    else:
        return origin, args


def table_rows_as_str(row_dicts: List[Dict[str, Any]], columns_per_row: int = 20):
    """
    Print a table row.

    :param row_dicts: The rows to print.
    :param columns_per_row: The maximum number of columns per row.
    """
    all_row_dicts_items = [list(row_dict.items()) for row_dict in row_dicts]
    # make items a list of n rows such that each row has a max size of 4
    all_items = [all_items[i:i + columns_per_row] for all_items in all_row_dicts_items
                 for i in range(0, len(all_items), columns_per_row)]
    keys = [list(map(lambda i: i[0], row)) for row in all_items]
    values = [list(map(lambda i: i[1], row)) for row in all_items]
    zipped_keys = list(zip(*keys))
    zipped_values = list(zip(*values))
    keys_values = [list(zip(zipped_keys[i], zipped_values[i])) for i in range(len(zipped_keys))]
    keys_values = [list(r[0]) + list(r[1]) if len(r) > 1 else r[0] for r in keys_values]
    all_table_rows = []
    row_values = [list(map(lambda v: str(v) if v is not None else "", row)) for row in keys_values]
    row_values = [list(map(lambda v: v[:150] + '...' if len(v) > 150 else v, row)) for row in row_values]
    row_values = [list(map(lambda v: v.lower() if v in ["True", "False"] else v, row)) for row in row_values]
    table = tabulate(row_values, tablefmt='simple_grid', maxcolwidths=[150] * 2)
    all_table_rows.append(table)
    return "\n".join(all_table_rows)


def row_to_dict(obj):
    return {
        col.name: getattr(obj, col.name)
        for col in obj.__table__.columns
        if not col.primary_key and not col.foreign_keys
    }


def dataclass_to_dict(obj):
    """
    Convert a dataclass to a dictionary.

    :param obj: The dataclass to convert.
    :return: The dictionary representation of the dataclass.
    """
    if is_dataclass(obj):
        return {f.name: getattr(obj, f.name) for f in fields(obj) if not f.name.startswith("_")}
    else:
        raise ValueError(f"Object {obj} is not a dataclass.")


def get_attribute_name(obj: Any, attribute: Optional[Any] = None, attribute_type: Optional[Type] = None,
                       possible_value: Optional[Any] = None) -> Optional[str]:
    """
    Get the name of an attribute from an object. The attribute can be given as a value, a type or a target value.
    And this method will try to find the attribute name using the given information.

    :param obj: The object to get the attribute name from.
    :param attribute: The attribute to get the name of.
    :param attribute_type: The type of the attribute to get the name of.
    :param possible_value: A possible value of the attribute to get the name of.
    :return: The name of the attribute.
    """
    attribute_name: Optional[str] = None
    if attribute_name is None and attribute is not None:
        attribute_name = get_attribute_name_from_value(obj, attribute)
    if attribute_name is None and attribute_type is not None:
        attribute_name = get_attribute_by_type(obj, attribute_type)[0]
    if attribute_name is None and possible_value is not None:
        attribute_name = get_attribute_by_type(obj, type(possible_value))[0]
    return attribute_name


def get_attribute_by_type(obj: Any, prop_type: Type) -> Tuple[Optional[str], Optional[Any]]:
    """
    Get a property from an object by type.

    :param obj: The object to get the property from.
    :param prop_type: The type of the property.
    """
    for name in dir(obj):
        if name.startswith("_") or callable(getattr(obj, name)):
            continue
        if isinstance(getattr(obj, name), (MetaData, registry)):
            continue
        prop_value = getattr(obj, name)
        if isinstance(prop_value, prop_type):
            return name, prop_value
        if hasattr(prop_value, "__iter__") and not isinstance(prop_value, str):
            if len(prop_value) > 0 and any(isinstance(v, prop_type) for v in prop_value):
                return name, prop_value
            else:
                # get args of type hint
                hint, origin, args = get_hint_for_attribute(name, obj)
                if origin in [list, set, tuple, dict, List, Set, Tuple, Dict]:
                    if prop_type is args[0]:
                        return name, prop_value
        else:
            # get the type hint of the attribute
            hint, origin, args = get_hint_for_attribute(name, obj)
            if hint is prop_type:
                return name, prop_value
            elif origin in [list, set, tuple, dict, List, Set, Tuple, Dict]:
                if prop_type is args[0]:
                    return name, prop_value
    return None, None


def get_attribute_name_from_value(obj: Any, attribute_value: Any) -> Optional[str]:
    """
    Get the name of an attribute from an object.

    :param obj: The object to get the attribute name from.
    :param attribute_value: The attribute value to get the name of.
    """
    for name in dir(obj):
        if name.startswith("_") or callable(getattr(obj, name)):
            continue
        prop_value = getattr(obj, name)
        if prop_value is attribute_value:
            return name


def get_attribute_values_transitively(obj: Any, attribute: Any) -> Any:
    """
    Get an attribute from a python object, if it is iterable, get the attribute values from all elements and unpack them
    into a list.

    :param obj: The object to get the sub attribute from.
    :param attribute: The  attribute to get.
    """
    if hasattr(obj, "__iter__") and not isinstance(obj, str):
        if isinstance(obj, (dict, UserDict)):
            all_values = [get_attribute_values_transitively(v, attribute) for v in obj.values()
                          if not isinstance(v, (str, type)) and hasattr(v, attribute)]
        else:
            all_values = [get_attribute_values_transitively(a, attribute) for a in obj
                          if not isinstance(a, (str, type)) and hasattr(a, attribute)]
        if can_be_a_set(all_values):
            return set().union(*all_values)
        else:
            return set(all_values)
    return getattr(obj, attribute)


def can_be_a_set(value: Any) -> bool:
    """
    Check if a value can be a set.

    :param value: The value to check.
    """
    if hasattr(value, "__iter__") and not isinstance(value, str):
        if len(value) > 0 and any(hasattr(v, "__iter__") and not isinstance(v, str) for v in value):
            return False
        else:
            return True
    else:
        return False


def get_all_subclasses(cls: Type) -> Dict[str, Type]:
    """
    Get all subclasses of a class recursively.

    :param cls: The class to get the subclasses of.
    :return: A dictionary of all subclasses.
    """
    all_subclasses: Dict[str, Type] = {}
    for sub_cls in cls.__subclasses__():
        all_subclasses[sub_cls.__name__.lower()] = sub_cls
        all_subclasses.update(get_all_subclasses(sub_cls))
    return all_subclasses


def make_set(value: Any) -> Set:
    """
    Make a set from a value.

    :param value: The value to make a set from.
    """
    return set(value) if is_iterable(value) else {value}


def make_value_or_raise_error(value: Any) -> Any:
    """
    Make a value or raise an error if the value is not a single value.

    :param value: The value to check.
    """
    if hasattr(value, "__iter__") and not isinstance(value, str):
        if hasattr(value, "__len__") and len(value) == 1:
            return list(value)[0]
        else:
            raise ValueError(f"Expected a single value, got {value}")
    return value


def tree_to_graph(root_node: Node) -> nx.DiGraph:
    """
    Convert anytree to a networkx graph.

    :param root_node: The root node of the tree.
    :return: A networkx graph.
    """
    graph = nx.DiGraph()
    unique_node_names = get_unique_node_names_func(root_node)

    def add_edges(node):
        if unique_node_names(node) not in graph.nodes:
            graph.add_node(unique_node_names(node))
        for child in node.children:
            if unique_node_names(child) not in graph.nodes:
                graph.add_node(unique_node_names(child))
            graph.add_edge(unique_node_names(node), unique_node_names(child), weight=child.weight)
            add_edges(child)

    add_edges(root_node)
    return graph


def get_unique_node_names_func(root_node) -> Callable[[Node], str]:
    nodes = [root_node]

    def get_all_nodes(node):
        for c in node.children:
            nodes.append(c)
            get_all_nodes(c)

    get_all_nodes(root_node)

    def nodenamefunc(node: Node):
        """
        Set the node name for the dot exporter.
        """
        similar_nodes = [n for n in nodes if n.name == node.name]
        node_idx = similar_nodes.index(node)
        return node.name if node_idx == 0 else f"{node.name}_{node_idx}"

    return nodenamefunc


def edge_attr_setter(parent, child):
    """
    Set the edge attributes for the dot exporter.
    """
    if child and hasattr(child, "weight") and child.weight:
        return f'style="bold", label=" {child.weight}"'
    return ""


def render_tree(root: Node, use_dot_exporter: bool = False,
                filename: str = "scrdr"):
    """
    Render the tree using the console and optionally export it to a dot file.

    :param root: The root node of the tree.
    :param use_dot_exporter: Whether to export the tree to a dot file.
    :param filename: The name of the file to export the tree to.
    """
    if not root:
        logging.warning("No rules to render")
        return
    for pre, _, node in RenderTree(root):
        print(f"{pre}{node.weight if hasattr(node, 'weight') and node.weight else ''} {node.__str__()}")
    if use_dot_exporter:
        unique_node_names = get_unique_node_names_func(root)

        de = DotExporter(root,
                         nodenamefunc=unique_node_names,
                         edgeattrfunc=edge_attr_setter
                         )
        de.to_dotfile(f"{filename}{'.dot'}")
        de.to_picture(f"{filename}{'.png'}")


def draw_tree(root: Node, fig: Figure):
    """
    Draw the tree using matplotlib and networkx.
    """
    # if matplotlib.get_backend().lower() not in ['qt5agg', 'qt4agg', 'qt6agg']:
    #     matplotlib.use("Qt6Agg")  # or "Qt6Agg", depending on availability

    if root is None:
        return
    fig.clf()
    graph = tree_to_graph(root)
    fig_sz_x = 13
    fig_sz_y = 10
    fig.set_size_inches(fig_sz_x, fig_sz_y)
    pos = nx.drawing.nx_agraph.graphviz_layout(graph, prog="dot")
    # scale down pos
    max_pos_x = max([v[0] for v in pos.values()])
    max_pos_y = max([v[1] for v in pos.values()])
    pos = {k: (v[0] * fig_sz_x / max_pos_x, v[1] * fig_sz_y / max_pos_y) for k, v in pos.items()}
    nx.draw(graph, pos, with_labels=True, node_color="lightblue", edge_color="gray", node_size=1000,
            ax=fig.gca(), node_shape="o", font_size=8)
    nx.draw_networkx_edge_labels(graph, pos, edge_labels=nx.get_edge_attributes(graph, 'weight'),
                                 ax=fig.gca(), rotate=False, clip_on=False)
    plt.pause(0.1)


def encapsulate_code_lines_into_a_function(code_lines: List[str], function_name: str, function_signature: str,
                                           func_doc: str, case_query: CaseQuery) -> str:
    """
    Encapsulate the given code lines into a function with the specified name, signature, and docstring.

    :param code_lines: The lines of code to include in the user input.
    :param function_name: The name of the function to include in the user input.
    :param function_signature: The function signature to include in the user input.
    :param func_doc: The function docstring to include in the user input.
    :param case_query: The case query object.
    """
    code = '\n'.join(code_lines)
    code = encapsulate_user_input(code, function_signature, func_doc)
    if case_query.is_function:
        args = "**case"
    else:
        args = "case"
    if f"return {function_name}({args})" not in code:
        code = code.strip() + f"\nreturn {function_name}({args})"
    return code



from __future__ import annotations

import os
import sys
from dataclasses import dataclass, field
from typing import List, Optional
from unittest import TestCase

try:
    from PyQt6.QtWidgets import QApplication
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
except ImportError as e:
    QApplication = None
    RDRCaseViewer = None

from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.experts import Human
from ripple_down_rules.helpers import is_matching
from ripple_down_rules.rdr import GeneralRDR


@dataclass
class WorldEntity:
    world: Optional[World] = field(default=None, kw_only=True, repr=False, hash=False)


@dataclass(unsafe_hash=True)
class Body(WorldEntity):
    name: str


@dataclass(unsafe_hash=True)
class Handle(Body):
    ...


@dataclass(unsafe_hash=True)
class Container(Body):
    ...


@dataclass(unsafe_hash=True)
class Connection(WorldEntity):
    parent: Body
    child: Body


@dataclass(unsafe_hash=True)
class FixedConnection(Connection):
    ...


@dataclass(unsafe_hash=True)
class PrismaticConnection(Connection):
    ...


@dataclass
class World:
    id: int = 0
    bodies: List[Body] = field(default_factory=list)
    connections: List[Connection] = field(default_factory=list)
    views: List[View] = field(default_factory=list, repr=False)

    def __eq__(self, other):
        if not isinstance(other, World):
            return False
        return self.id == other.id


@dataclass(unsafe_hash=True)
class View(WorldEntity):
    ...


@dataclass(unsafe_hash=True)
class Drawer(View):
    handle: Handle
    container: Container
    correct: Optional[bool] = None


@dataclass
class Cabinet(View):
    container: Container
    drawers: List[Drawer] = field(default_factory=list)

    def __hash__(self):
        return hash((self.__class__.__name__, self.container))


class TestRDRWorld(TestCase):
    drawer_case_queries: List[CaseQuery]
    world: World
    app: Optional[QApplication] = None
    viewer: Optional[RDRCaseViewer] = None

    @classmethod
    def setUpClass(cls):
        world = World()
        cls.world = world

        handle = Handle('h1', world=world)
        handle_2 = Handle('h2', world=world)
        container_1 = Container('c1', world=world)
        container_2 = Container('c2', world=world)
        connection_1 = FixedConnection(container_1, handle, world=world)
        connection_2 = PrismaticConnection(container_2, container_1, world=world)

        world.bodies = [handle, container_1, container_2, handle_2]
        world.connections = [connection_1, connection_2]

        all_possible_drawers = []
        for handle in [body for body in world.bodies if isinstance(body, Handle)]:
            for container in [body for body in world.bodies if isinstance(body, Container)]:
                view = Drawer(handle, container, world=world)
                all_possible_drawers.append(view)

        # print(all_possible_drawers)
        cls.drawer_case_queries = [CaseQuery(possible_drawer, "correct", (bool,), True, default_value=False)
                                   for possible_drawer in all_possible_drawers]

        if RDRCaseViewer is not None and QApplication is not None:
            cls.app = QApplication(sys.argv)
            cls.viewer = RDRCaseViewer(save_file="./test_generated_rdrs")

    def test_view_rdr(self):
        self.get_view_rdr(use_loaded_answers=True, save_answers=False, append=False)

    def test_save_and_load_view_rdr(self):
        view_rdr = self.get_view_rdr(use_loaded_answers=True, save_answers=False, append=False)
        filename = os.path.join(os.getcwd(), "test_results/world_views_rdr")
        model_name = view_rdr.save(filename)
        loaded_rdr = GeneralRDR.load(filename, model_name=model_name)
        self.assertEqual(view_rdr.classify(self.world), loaded_rdr.classify(self.world))
        self.assertEqual(self.world.bodies, loaded_rdr.start_rules[0].corner_case.bodies)

    def test_write_view_rdr_to_python_file(self):
        rdrs_dir = "./test_generated_rdrs/view_rdr"
        os.makedirs(rdrs_dir, exist_ok=True)
        view_rdr = self.get_view_rdr()
        view_rdr._write_to_python(rdrs_dir)
        loaded_rdr_classifier = view_rdr.get_rdr_classifier_from_python_file(rdrs_dir)
        found_views = loaded_rdr_classifier(self.world)
        self.assertTrue(len([v for v in found_views["views"] if isinstance(v, Drawer)]) == 1)
        self.assertTrue(len([v for v in found_views["views"] if isinstance(v, Cabinet)]) == 1)
        self.assertTrue(len(found_views["views"]) == 2)

    def get_view_rdr(self, views=(Drawer, Cabinet), use_loaded_answers: bool = True, save_answers: bool = False,
                     append: bool = False):
        expert = Human(use_loaded_answers=use_loaded_answers, append=append, viewer=self.viewer)
        filename = os.path.join(os.getcwd(), "test_expert_answers/view_rdr_expert_answers_fit")
        if use_loaded_answers:
            expert.load_answers(filename)
        rdr = GeneralRDR()
        for view in views:
            rdr.fit_case(CaseQuery(self.world, "views", (view,), False), expert=expert)
        if save_answers:
            expert.save_answers(filename)

        found_views = rdr.classify(self.world)
        # print(found_views)
        for view in views:
            self.assertTrue(len([v for v in found_views["views"] if isinstance(v, view)]) > 0)

        return rdr

    def test_drawer_rdr(self):
        self.get_drawer_rdr(use_loaded_answers=True, save_answers=False)

    def test_write_drawer_rdr_to_python_file(self):
        rdrs_dir = "./test_generated_rdrs/drawer_rdr"
        os.makedirs(rdrs_dir, exist_ok=True)
        drawer_rdr = self.get_drawer_rdr()
        drawer_rdr._write_to_python(rdrs_dir)
        loaded_rdr_classifier = drawer_rdr.get_rdr_classifier_from_python_file(rdrs_dir)
        for case_query in self.drawer_case_queries:
            self.assertTrue(is_matching(loaded_rdr_classifier, case_query))

    def get_drawer_rdr(self, use_loaded_answers: bool = True, save_answers: bool = False):
        expert = Human(use_loaded_answers=use_loaded_answers)
        filename = os.path.join(os.getcwd(), "test_expert_answers/correct_drawer_rdr_expert_answers_fit")
        if use_loaded_answers:
            expert.load_answers(filename)
        rdr = GeneralRDR()
        rdr.fit(self.drawer_case_queries, expert=expert, animate_tree=False)
        if save_answers:
            expert.save_answers(filename)
        for case_query in self.drawer_case_queries:
            self.assertTrue(is_matching(rdr.classify, case_query))
        return rdr



from IPython.core.interactiveshell import ExecutionInfo
from IPython.terminal.embed import InteractiveShellEmbed
from traitlets.config import Config

from ripple_down_rules.utils import capture_variable_assignment, contains_return_statement, extract_dependencies


class IpythonShell(InteractiveShellEmbed):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.all_lines = []

    def run_cell(self, raw_cell: str, **kwargs):
        """
        Override the run_cell method to capture return statements.
        """
        if contains_return_statement(raw_cell):
            self.all_lines.append(raw_cell)
            print("Exiting shell on `return` statement.")
            self.ask_exit()
            return None
        result = super().run_cell(raw_cell, **kwargs)
        if not result.error_in_exec:
            self.all_lines.append(raw_cell)
        return result


class IpythonShellManager:
    def __init__(self, scope=None, header=None):
        self.scope = scope or {}
        self.header = header or ">>> Embedded Ipython Shell"
        self.raw_condition = None
        self.shell = self._init_shell()
        self.all_code_lines = []

    def _init_shell(self):
        """
        Initialize the Ipython shell with a custom configuration.
        """
        cfg = Config()
        shell = IpythonShell(config=cfg, user_ns=self.scope, banner1=self.header)
        return shell

    def run(self):
        """
        Run the embedded shell.
        """
        self.shell()
        self.all_code_lines = extract_dependencies(self.shell.all_lines)
        user_input = f"def get_value(case):\n    "
        user_input += '\n    '.join(self.all_code_lines)
        print(user_input)
        eval(compile(user_input, '<string>', 'exec'), self.scope)
        print(self.scope['get_value'](4))


def run_ipython_shell():
    x = 10
    case = 5
    scope = locals()
    mgr = IpythonShellManager(scope)
    mgr.run()
    print(mgr.all_code_lines)
    # Apply changes to outer scope
    x = scope['x']
    print(f"Back in code: x={x}")


# run_ipython_shell()



from __future__ import annotations

import ast
import json
import logging
import os
import uuid
from abc import ABC, abstractmethod

from typing_extensions import Optional, TYPE_CHECKING, List

from .datastructures.callable_expression import CallableExpression
from .datastructures.enums import PromptFor
from .datastructures.dataclasses import CaseQuery
from .datastructures.case import show_current_and_corner_cases
from .utils import extract_imports, extract_function_source, get_imports_from_scope, encapsulate_user_input

try:
    from .user_interface.gui import RDRCaseViewer
except ImportError as e:
    RDRCaseViewer = None
from .user_interface.prompt import UserPrompt

if TYPE_CHECKING:
    from .rdr import Rule


class Expert(ABC):
    """
    The Abstract Expert class, all experts should inherit from this class.
    An expert is a class that can provide differentiating features and conclusions for a case when asked.
    The expert can compare a case with a corner case and provide the differentiating features and can also
    provide one or multiple conclusions for a case.
    """
    all_expert_answers: Optional[List] = None
    """
    A list of all expert answers, used for testing purposes.
    """
    use_loaded_answers: bool = False
    """
    A flag to indicate if the expert should use loaded answers or not.
    """

    def __init__(self, use_loaded_answers: bool = True,
                 append: bool = False,
                 answers_save_path: Optional[str] = None):
        self.all_expert_answers = []
        self.use_loaded_answers = use_loaded_answers
        self.append = append
        self.answers_save_path = answers_save_path
        if answers_save_path is not None:
            if use_loaded_answers:
                self.load_answers(answers_save_path)
            else:
                os.remove(answers_save_path + '.py')
            self.append = True

    @abstractmethod
    def ask_for_conditions(self, case_query: CaseQuery, last_evaluated_rule: Optional[Rule] = None) \
            -> CallableExpression:
        """
        Ask the expert to provide the differentiating features between two cases or unique features for a case
        that doesn't have a corner case to compare to.

        :param case_query: The case query containing the case to classify and the required target.
        :param last_evaluated_rule: The last evaluated rule.
        :return: The differentiating features as new rule conditions.
        """
        pass

    @abstractmethod
    def ask_for_conclusion(self, case_query: CaseQuery) -> Optional[CallableExpression]:
        """
        Ask the expert to provide a relational conclusion for the case.

        :param case_query: The case query containing the case to find a conclusion for.
        :return: A callable expression that can be called with a new case as an argument.
        """

    def clear_answers(self, path: Optional[str] = None):
        """
        Clear the expert answers.

        :param path: The path to clear the answers from. If None, the answers will be cleared from the
                     answers_save_path attribute.
        """
        if path is None and self.answers_save_path is None:
            raise ValueError("No path provided to clear expert answers, either provide a path or set the "
                                "answers_save_path attribute.")
        if path is None:
            path = self.answers_save_path
        if os.path.exists(path + '.json'):
            os.remove(path + '.json')
        if os.path.exists(path + '.py'):
            os.remove(path + '.py')
        self.all_expert_answers = []

    def save_answers(self, path: Optional[str] = None):
        """
        Save the expert answers to a file.

        :param path: The path to save the answers to.
        """
        if path is None and self.answers_save_path is None:
            raise ValueError("No path provided to save expert answers, either provide a path or set the "
                                "answers_save_path attribute.")
        if path is None:
            path = self.answers_save_path
        is_json = os.path.exists(path + '.json')
        if is_json:
            self._save_to_json(path)
        else:
            self._save_to_python(path)

    def _save_to_json(self, path: str):
        """
        Save the expert answers to a JSON file.

        :param path: The path to save the answers to.
        """
        all_answers = self.all_expert_answers
        if self.append and os.path.exists(path + '.json'):
            # read the file and append the new answers
            with open(path + '.json', "r") as f:
                old_answers = json.load(f)
                all_answers = old_answers + all_answers
        with open(path + '.json', "w") as f:
            json.dump(all_answers, f)

    def _save_to_python(self, path: str):
        """
        Save the expert answers to a Python file.

        :param path: The path to save the answers to.
        """
        dir_name = os.path.dirname(path)
        if not os.path.exists(dir_name + '/__init__.py'):
            os.makedirs(dir_name, exist_ok=True)
            with open(dir_name + '/__init__.py', 'w') as f:
                f.write('# This is an empty init file to make the directory a package.\n')
        action = 'w' if not self.append else 'a'
        with open(path + '.py', action) as f:
            for scope, func_source in self.all_expert_answers:
                if len(scope) > 0:
                    imports = '\n'.join(get_imports_from_scope(scope)) + '\n\n\n'
                else:
                    imports = ''
                if func_source is not None:
                    uid = uuid.uuid4().hex
                    func_source = encapsulate_user_input(func_source, CallableExpression.get_encapsulating_function(f'_{uid}'))
                else:
                    func_source = 'pass  # No user input provided for this case.\n'
                f.write(imports + func_source + '\n' + '\n\n\n\'===New Answer===\'\n\n\n')

    def load_answers(self, path: Optional[str] = None):
        """
        Load the expert answers from a file.

        :param path: The path to load the answers from.
        """
        if path is None and self.answers_save_path is None:
            raise ValueError("No path provided to load expert answers from, either provide a path or set the "
                                "answers_save_path attribute.")
        if path is None:
            path = self.answers_save_path
        is_json = os.path.exists(path + '.json')
        if is_json:
            self._load_answers_from_json(path)
        elif os.path.exists(path + '.py'):
            self._load_answers_from_python(path)

    def _load_answers_from_json(self, path: str):
        """
        Load the expert answers from a JSON file.

        :param path: The path to load the answers from.
        """
        with open(path + '.json', "r") as f:
            all_answers = json.load(f)
        self.all_expert_answers = [({}, answer) for answer in all_answers]

    def _load_answers_from_python(self, path: str):
        """
        Load the expert answers from a Python file.

        :param path: The path to load the answers from.
        """
        file_path = path + '.py'
        with open(file_path, "r") as f:
            all_answers = f.read().split('\n\n\n\'===New Answer===\'\n\n\n')[:-1]
        all_function_sources = list(extract_function_source(file_path, []).values())
        all_function_sources_names = list(extract_function_source(file_path, []).keys())
        for i, answer in enumerate(all_answers):
            answer = answer.strip('\n').strip()
            if 'def ' not in answer and 'pass' in answer:
                self.all_expert_answers.append(({}, None))
                continue
            scope = extract_imports(tree=ast.parse(answer))
            function_source = all_function_sources[i].replace(all_function_sources_names[i],
                                                              CallableExpression.encapsulating_function_name)
            self.all_expert_answers.append((scope, function_source))


class Human(Expert):
    """
    The Human Expert class, an expert that asks the human to provide differentiating features and conclusions.
    """

    def __init__(self, viewer: Optional[RDRCaseViewer] = None, **kwargs):
        """
        Initialize the Human expert.

        :param viewer: The RDRCaseViewer instance to use for prompting the user.
        """
        super().__init__(**kwargs)
        self.user_prompt = UserPrompt(viewer)

    def ask_for_conditions(self, case_query: CaseQuery,
                           last_evaluated_rule: Optional[Rule] = None) \
            -> CallableExpression:
        if (not self.use_loaded_answers or len(self.all_expert_answers) == 0) and self.user_prompt.viewer is None:
            show_current_and_corner_cases(case_query.case, {case_query.attribute_name: case_query.target_value},
                                          last_evaluated_rule=last_evaluated_rule)
        return self._get_conditions(case_query)

    def _get_conditions(self, case_query: CaseQuery) \
            -> CallableExpression:
        """
        Ask the expert to provide the differentiating features between two cases or unique features for a case
        that doesn't have a corner case to compare to.

        :param case_query: The case query containing the case to classify.
        :return: The differentiating features as new rule conditions.
        """
        user_input = None
        if self.use_loaded_answers and len(self.all_expert_answers) == 0 and self.append:
            self.use_loaded_answers = False
        if self.use_loaded_answers:
            try:
                loaded_scope, user_input = self.all_expert_answers.pop(0)
            except IndexError:
                self.use_loaded_answers = False
        if user_input is not None:
            case_query.scope.update(loaded_scope)
            condition = CallableExpression(user_input, bool, scope=case_query.scope)
        else:
            user_input, condition = self.user_prompt.prompt_user_for_expression(case_query, PromptFor.Conditions)
        if not self.use_loaded_answers:
            self.all_expert_answers.append((condition.scope, user_input))
            if self.answers_save_path is not None:
                self.save_answers()
        case_query.conditions = condition
        return condition

    def ask_for_conclusion(self, case_query: CaseQuery) -> Optional[CallableExpression]:
        """
        Ask the expert to provide a conclusion for the case.

        :param case_query: The case query containing the case to find a conclusion for.
        :return: The conclusion for the case as a callable expression.
        """
        expression: Optional[CallableExpression] = None
        expert_input: Optional[str] = None
        if self.use_loaded_answers and len(self.all_expert_answers) == 0 and self.append:
            self.use_loaded_answers = False
        if self.use_loaded_answers:
            try:
                loaded_scope, expert_input = self.all_expert_answers.pop(0)
                if expert_input is not None:
                    case_query.scope.update(loaded_scope)
                    expression = CallableExpression(expert_input, case_query.attribute_type,
                                                    scope=case_query.scope,
                                                    mutually_exclusive=case_query.mutually_exclusive)
            except IndexError:
                self.use_loaded_answers = False
        if not self.use_loaded_answers:
            if self.user_prompt.viewer is None:
                show_current_and_corner_cases(case_query.case)
            expert_input, expression = self.user_prompt.prompt_user_for_expression(case_query, PromptFor.Conclusion)
            if expression is None:
                self.all_expert_answers.append(({}, None))
            else:
                self.all_expert_answers.append((expression.scope, expert_input))
            if self.answers_save_path is not None:
                self.save_answers()
        case_query.target = expression
        return expression
